# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Create the link between freeart and a configuration file. Basically set up
a reconstruction
"""
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/10/2017"

from freeart.configuration import config, structs
import freeart
import os
import sys
import freeart.utils
import freeart.utils.reconstrutils as reconstrutils
import freeart.utils.physicalelmts as physicalelmts
from freeart.utils import sinogramselection
import numpy as np
import logging
import fisx
from threading import Thread
from fisx import Detector
from fabio.edfimage import edfimage
from freeart.configuration import read, fileInfo
import numpy
import copy
_logger = logging.getLogger("freeARTInterpreter")
logging.basicConfig(level=logging.INFO)

elementInstance = None


class AbsConfInterpreter(object):
    """Abstract class from which each interpreter (fluorescence, tx...)
    should inherit"""

    def __init__(self, filePath, _config):
        assert not (filePath is None and _config is None)

        self.filePath = filePath
        if filePath:
            self.config = read(self.filePath)
        else:
            self.config = copy.deepcopy(_config)

        assert(self.config.precision in config._ReconsConfig.PRECISION_TO_TYPE)
        self.precisionType = config._ReconsConfig.PRECISION_TO_TYPE[self.config.precision]
        self.setTheoreticalSampSize()
        self.setLimitParallelReconstruction(-1)
        self.setLimitVoxels(-1)
        self.config.fillSinogramNames()
        self.angles = None
        if self.config.useAFileForI0:
            self.config.I0.loadData()
            assert self.config.I0.data is not None
            self.config.I0.data = self.adaptSinogramData(data=self.config.I0.data,
                                                         isI0=True)

    def resetAngles(self):
        """Reset all angles """
        self.angles = None

    def setTheoreticalSampSize(self):
        raise NotImplementedError('AbsConfInterpreter is a pure virtual class')

    def iterate(self):
        """
        Iterate on the reconstruction. Each iteration generate a backward
        projection for each angle of acquisition (in a random order).
        """
        raise NotImplementedError('AbsConfInterpreter is a pure virtual class')

    def _reduceSinogram(self, matrix):
        """Reduce the data according to the DefinitionReductionfactor and the
        ProjectionReductionFactor

        :param matrix: the matrix to reduce
        :return: the matrix reduced
        """
        if self.config.definitionReduction is None:
            self.config.definitionReduction = 1

        matrix = self._reduceXData(matrix, self.config.definitionReduction)

        if self.config.projectionNumberReduction is None:
            self.config.projectionNumberReduction = 1

        return self._reduceNbAnglesData(matrix)

    def _adaptMatToReduction(self, data):
        """
        Reduce the matrix in x and y according to the requested configuration.
        Take into account the projection reduction and the definition reduction.

        :param data: the matrix to reduce.
        :return: a reduced matrix
        """
        if data.shape[1] != self._getTheoreticalSampSize():
            if self.config.definitionReduction is None:
                self.dataXReduction = self.config.definitionReduction
            res = None
            reduction = self.config.definitionReduction
            if len(data.shape) == 3:
                res = data[:, ::reduction, ::reduction]

            if len(data.shape) == 2:
                res = data[::reduction, ::reduction]

            if res is None:
                raise RuntimeError("Unvalid matrix dimension")
            else:
                return res
        else:
            return data

    def _getTheoreticalSampSize(self):
        # return the size of the reconstructed sample
        return self._thSampSize

    def _getFirstFluoSino(self):
        if self.config.sinograms is None or len(self.config.sinograms) is 0:
            return None
        else:
            return self.config.sinograms[0]

    def _getAngles(self, nbAngles, selection):
        """
        Compute nbAngle homogeneous angle between minAngle and maxAngle.
        Take into account if the last angle should be equal to the first angle
        from the config file (LastAngleEqualFirstAngle option)

        :param nbAngles: the number of angle we want to generate
        :return: the angle for homogeneous angle between each acquisition
        """
        def getMinAngle():
            if self.config.minAngle is None:
                err = "can't find information about min angle in the .cfg file"
                raise IOError(err)

            return self.config.minAngle

        def getMaxAngle():
            if self.config.maxAngle is None:
                err = "can't find information about max angle in the .cfg file"
                raise IOError(err)

            return self.config.maxAngle

        # if specify that the last angle in the sinogram == first angle in the
        # sinogram
        if self.config.includeLastProjection is None:
            self.config.includeLastProjection = False

        minAngle = getMinAngle()
        maxAngle = getMaxAngle()

        if minAngle > maxAngle:
            raise RuntimeError("minAngle > maxAngle, can't process")

        if minAngle < -2.1*np.pi or minAngle > 2.1*np.pi or maxAngle < -2.1*np.pi or maxAngle > 2.1*np.pi:
            err = """Angles should be defined in radians. We for minAngle and
                maxAngle at least one angle < -2*pi or > 2.pi """
            raise RuntimeError(err)

        return sinogramselection.getAngles(minAngle=minAngle,
                                           maxAngle=maxAngle,
                                           selection=selection,
                                           acquiInverted=self.config.acquiInverted,
                                           lastAngleEqualFirst=self.config.includeLastProjection,
                                           nbAngles=nbAngles)

    def _reduceNbAnglesData(self, data):
        """
        Reduce the number of angles in the data of a factor factorReduction

        :param data: the matrix to reduce
        :param factorReduction: the factor of reduction to apply
        :return: the matrix reduced

        .. warning:: the matrix should be a 2D numpy array
        """
        return data[::self.config.projectionNumberReduction, ]

    def _reduceXData(self, data, factorReduction):
        """
        Reduce the number of projection in the data of a factor factorReduction

        :param data: the matrix to reduce
        :param factorReduction: the factor of reduction to apply
        :return: the matrix reduced

        .. warning:: the matrix should be a 2D numpy array
        """
        return data[:, ::factorReduction]

    def _centerSinogram(self, data, center):
        """Simple function to center the sinogram

        :param data: the sinogram to center
        :param center: The new center of rotation
        """
        assert(len(data.shape) == 2)
        width = min(abs(data.shape[1]-center), center)
        if width > 0:
            return data[:, center-width:center+width]
        else:
            _logger.info("can t center the sinogram. Centering avoid")
            return data

    def getSinograms(self):
        """

        :return: A dictionnary of the sinograms to treat"""
        err = 'getSinograms not implemented. '
        err += 'AbsConfInterpreter is a pure virtual class'
        raise NotImplementedError(err)

    def getFile(self):
        """

        :return: the current configuratioin file """
        return self.filePath

    def setLimitParallelReconstruction(self, val):
        """
        Limit the number of parallel reconstruction we can run

        :param val: the new limitation of parallel reconstructions to run
        """
        if val > 0:
            self.limitParaRecon = val
        if val == -1:
            self.limitParaRecon = sys.maxsize

    def setLimitVoxels(self, val):
        """
        Limit the maximal number of voxel we can run during one iteration

        :param val: the maximal number of voxel to run during one step.
        """
        if val > 0:
            self.limitVoxel = val
        if val == -1:
            self.limitVoxel = sys.maxsize

    def _normalizeRawData(self, sinogram):
        """Normalization of raw data. Apply the new center of rotation, reduce
        data if needed and normalize fron I0 value.

        :param sinogram: the sinogram to normalize
        """
        assert sinogram.ndim is 2
        if self.config.useAFileForI0 is True:
            assert self.config.I0.data is not None
            sI0 = reconstrutils.decreaseMatSize(self.config.I0.data).shape
            if (sI0 != sinogram.shape):
                err = """I0 sinogram (%s) and sinogram (%s) have different
                                    shapes""" % (sI0, sinogram.shape)
                raise ValueError(err)
            else:
                sinogram = sinogram / reconstrutils.decreaseMatSize(self.config.I0.data)
        elif self.config.I0 is not None:
            if not isinstance(self.config.I0, numpy.ndarray) and self.config.I0 <= 0.0:
                raise ValueError("invalid I0 value given")
            else:
                sinogram = sinogram / self.config.I0

        return sinogram

    def adaptSinogramData(self, data, isI0=False):
        """
        center, reduce sinogram (X definition and number of projection) and
        pick the range of projection requested.

        :param data:
        :param computeAngles: If true then compute `self.angles`
        :return: data adapted to constrain
        """
        data = reconstrutils.decreaseMatSize(data)

        # step 0: get angles
        selection = self.config.projections
        if isI0 is False:
            self.angles = self._getAngles(nbAngles=data.shape[0],
                                          selection=selection)
            self.angles = self._reduceNbAnglesData(self.angles)
            self.angles = numpy.ascontiguousarray(self.angles)

        # step 1 : normalization of the sinogram
        if self.config.center:
            data = self._centerSinogram(data, self.config.center)

        # step 2: get the selected sinogram
        data = sinogramselection.getSelection(selection=selection,
                                              projections=data)

        # step 3 : deal with data reduction.
        # Rule : do not reduce data when the target is to produce an absMat.
        # Simplify a lot stuff (specially when loading fluo config file)
        data = self._reduceSinogram(data)

        # step 4 : normalize from I0
        if isI0 is False:
            data = self._normalizeRawData(data)

        # step 5 : move it to 3D
        if data.ndim is 2:
            data.shape = (1,
                          data.shape[0],
                          data.shape[1])
        return data

    def getSinograms(self):
        """

        :return: A dictionnary of the sinograms to treat
        """
        return self.config.sinograms


class TxConfInterpreter(AbsConfInterpreter):
    """
    Interpreter for tx reconstruction

    :param filePath: the path of the cfg file
    """
    def __init__(self, filePath, _config=None):
        AbsConfInterpreter.__init__(self, filePath=filePath, _config=_config)
        if _config:
            assert(isinstance(_config, config.TxConfig))
        self.txRecons = {}
        # make sure data is here
        self._setUpTxReconstruction()

    def getReconstructionAlgorithms(self):
        """

        :return: the dictionay of the freeart reconstruction.
        """
        return self.txRecons

    def setTheoreticalSampSize(self):
        def checkSinograms():
            shape = None
            for sinogram in self.config.sinograms:
                assert isinstance(sinogram, structs.Sinogram)
                assert (sinogram.data is not None)
                assert (sinogram.data.ndim is 2)

                if shape is None:
                    shape = sinogram.data.shape
                elif sinogram.data.shape != shape:
                    raise ValueError('At least two sinograms to reconstruct'
                                     ' have different dimension. Case not'
                                     ' managed by the freeart interpreter')

        # TODO: should be removed
        assert(self.config.sinograms)
        for sinogram in self.config.sinograms:
            sinogram.loadData()

        checkSinograms()
        width = sinogram.data.shape[1]
        if self.config.center not in (None, -1):
            hwidth = self.config.center if self.config.center <= width / 2 else width - self.config.center
            width = hwidth * 2
        mod = width % self.config.definitionReduction
        mod = 1 if mod > 0 else 0
        self._thSampSize = int(width / self.config.definitionReduction) + mod

    def _setUpTxReconstruction(self):
        """
        Setup the FreeART Tx algorithm to fit with the self.configuration file
        """
        # deal with I0
        if self.config.useAFileForI0 is True:
            assert(isinstance(self.config.I0, structs.I0Sinogram))
            if self.config.I0.data is None:
                self.config.I0.loadData()

        for sinogram in self.config.sinograms:
            self.txRecons[sinogram.name] = self._buildAlgoTx(sinogram=sinogram)

    def _buildAlgoTx(self, sinogram):
        """
        Build a transmission ART algorithm for the given algorithm and given
        self.configuration file
        """
        assert sinogram.data is not None
        sinogram.data = self.adaptSinogramData(sinogram.data,
                                               isI0=False)

        # if some values of the sinogram are negative and the sinogram is I
        # (and not -log(I))
        if self.config.computeLog and (True in (sinogram.data < 0.0)):
            err = """given sinogram has negative values. Can't process
                  the reconstruction"""
            raise RuntimeError(err)

        # create the adapted ARTAlgorithm
        if self.config.computeLog is True:
            logVal = -np.log(sinogram.data)
        else:
            logVal = sinogram.data
        assert(np.nan not in logVal)

        txAlgo = freeart.TxBckProjection(logVal.astype(self.precisionType),
                                         self.angles)

        if self.config.voxelSize:
            txAlgo.setVoxelSize(self.config.voxelSize * self.config.definitionReduction)

        if self.config.oversampling:
            txAlgo.setOverSampling(self.config.oversampling)

        if self.config.I0 is not None:
            if self.config.useAFileForI0 is False:
                txAlgo.setI0(float(self.config.I0))

        if self.config.beamCalcMethod:
            txAlgo.setRayPointCalculationMethod(self.config.beamCalcMethod)

        if self.config.dampingFactor:
            txAlgo.setDampingFactor(self.config.dampingFactor)

        return txAlgo

    def iterate(self, nbIteration):
        """
        Iterate on the reconstruction. Each iteration generate a backward
        projection for each angle of acquisition (in a random order).
        """
        if self.txRecons is None:
            err = "No ART algorithm defined"
            raise ValueError(err)
        threads = []
        for algoID in self.txRecons:
            assert(not self.txRecons[algoID] is None)
            threads.append(Thread(target=self.txRecons[algoID].iterate,
                                  args=(nbIteration,)))

        [thread.start() for thread in threads]
        [thread.join() for thread in threads]

    def setRandSeedToZero(self, b):
        if self.txRecons is None:
            err = "No ART algorithm defined"
            raise ValueError(err)
        for algo in self.txRecons:
            self.txRecons[algo].setRandSeedToZero(b)


class FluoConfInterpreter(AbsConfInterpreter):
    """
    FreeART interpreter for fluorescence reconstruction.
    Basicall y take a .cfg file containing the description of the
    reconstruction in input and create all corresponding ARTAlgorithm for such
    reconstructions

    :param filePath: the configuration file to be interpreted.
    """
    def __init__(self, filePath, _config=None):
        AbsConfInterpreter.__init__(self, filePath=filePath, _config=_config)

        if _config:
            assert (isinstance(_config, config.FluoConfig))

        self.fisxMaterials = {}
        """Dict to associate at each material in the sample the corresponding
        fisx materials"""
        self.selfAbsMatFile = {}
        """File where to get the selfAbsMat file in case the user give them"""
        self.absMatrixFile = None
        """The file of the absorption matrix file"""
        self.fluoSinogramsAlgorithms = {}
        "all the art algorithm to fit the configuration file"
        self.fluoSinograms = {}
        "sinograms to treat"
        self.interactionMatrice = {}
        "the interaction matrices"
        self._sinoElemt = {}

        self._setUpFluoReconstruction()
        self._buildAlgorithms()

    def setTheoreticalSampSize(self):
        sino = self._getFirstFluoSino()
        sino.loadData()
        assert sino.data is not None
        assert sino
        assert isinstance(sino, structs.FluoSino)
        assert sino.data.ndim is 2

        width = sino.data.shape[1]
        if self.config.center not in (None, -1):
            hwidth = self.config.center if self.config.center <= width / 2 else width - self.config.center
            width = hwidth * 2
        mod = width % self.config.definitionReduction
        mod = 1 if mod > 0 else 0

        self._thSampSize = int(width / self.config.definitionReduction) + mod

    def _setUpFluoReconstruction(self):
        """
        Setup the FreeART Fluo algorithm to fit with the self.configuration
        file
        """
        if len(self.config.sinograms) < 1:
            raise ValueError("No fluorescence sinograms given")

        if self.config.absMat is None or (self.config.absMat.fileInfo is None and self.config.absMat.data is None):
            raise ValueError('Absorption matrix not defined')

        if self.config.absMat.data is None and not os.path.isfile(self.config.absMat.fileInfo.getFile()):
            err = self.config.absMat + " absorption file setted does not exist"
            raise ValueError(err)

        # in the case we need to build the absorption matrice from ourself
        if self.config.isAbsMatASinogram is True:
            mess = """The interpreter is not able to reconstruct by himself the
                absorption matrix. Please build it first, then give if back to
                the freeart interpreter"""
            raise RuntimeError(mess)
        elif self.config.absMat.data is None:
            self.config.absMat.loadData()
        self.config.absMat.data = self._adaptMatToReduction(self.config.absMat.data)
        if self.config.absMat.data.shape[1] != self._getTheoreticalSampSize():
            s1 = self.config.absMat.data.shape[1]
            s2 = self._getTheoreticalSampSize()
            err = 'The absorption matrix has invalid dimensions (%s).' \
                  'When in theory it should be (%s, %s). This error can occur' \
                  'if the absorption matrix have been reconstructed with a ' \
                  'different center of rotation for example.' % (s1, s2, s2)
            raise ValueError(err)

        self.config.absMat.data.shape = (self.config.absMat.data.shape[0],
                                         self.config.absMat.data.shape[1],
                                         1)

        self.__prepareMaterials()

    def __prepareMaterials(self):
        """
        prepare materials for fisx
        """
        if (self.config.materials is None or
            self.config.materials.materials is None or
            self.config.materials.matComposition is None or
            self.config.materials.materials.fileInfo is None or
            self.config.materials.matComposition.fileInfo is None or
            fileInfo.FreeARTFileInfo._equalOrNone(self.config.materials.materials.fileInfo, None) or
            fileInfo.FreeARTFileInfo._equalOrNone(self.config.materials.matComposition.fileInfo, None)):
            return
        else:
            assert(self.config.materials.materials is not None)
            assert(self.config.materials.materials.fileInfo is not None)
            assert(self.config.materials.matComposition is not None)
            assert(self.config.materials.matComposition.fileInfo is not None)
            assert(os.path.isfile(self.config.materials.materials.fileInfo.file_path()))
            self.config.materials.loadData()
            materials = self.config.materials.materials.data

            physicalelmts.elementInstance.removeMaterials()
            for materialName in materials:
                uniqMatName = reconstrutils.getFisxMatName(materialName)
                self.fisxMaterials[uniqMatName] = reconstrutils.convertToFisxMaterial(
                    uniqMatName,
                    materials[materialName])
                physicalelmts.elementInstance.addMaterial(
                    self.fisxMaterials[uniqMatName])

    def __getFisxDetector(self):
        """

        :return: the fisx detector to compute the interaction matrix
        """
        # TODO : get the material of the detector
        # pb : in this case we will take twice the effect of the solid angle.
        # Something that we don't want to...
        detector = Detector("Si1", 2.33, 1.0)
        detector.setDiameter(self.config.detector.width)

    @staticmethod
    def getProbabilityOfEmission(fisxMat, energy, element, fisxElements,
                                 shell=None):
        """
        compute the probability of emission of the element for a specific
        energy and material.

        :param fisxMat: the fisx material of the interaction
        :param energy: the nergy of the incomng beam
        :param element: the generated element
        :param shell: the targetted shell of the interaction.

        :return: the probability of emission of the element for a specific
                 energy and material.
        """
        if not type(fisxMat) is fisx.Material:
            raise TypeError('material should be a fisx.Material')
        if(len(fisxMat.getComposition()) is 0):
            raise TypeError('Issue with fisx Material, not correctly registred')

        excitations = fisxElements.getExcitationFactors(element, [energy])
        compo = fisxMat.getComposition()
        compo_compa = {}
        for mat in compo:
            if sys.version_info[1] > 2 and type(mat) is str:
                compo_compa[mat.encode('utf-8')] = compo[mat]
            else:
                compo_compa[mat] = compo[mat]

        if sys.version_info[1] > 2:
            photonelectricMassAttenuation = fisxElements.getMassAttenuationCoefficients(compo_compa, [energy])[b'photoelectric']
        else:
            photonelectricMassAttenuation = fisxElements.getMassAttenuationCoefficients(compo_compa, [energy])['photoelectric']

        if shell is None:
            rate = 0.
            for excitation in excitations:
                rate += excitations[excitation]['rate'] * photonelectricMassAttenuation[0]
            return rate
        else:
            if shell not in excitations:
                raise ValueError('Required shell is not existing for the elemnt')
            else:
                return excitations[element]['rate'][shell] * photonelectricMassAttenuation

    def _resetFluoSinograms(self):
        """Reset all fluorescence sinograms """
        self.fluoSinogramsAlgorithms = {}
        self.fluoSinograms = {}
        self.interactionMatrice = {}

    def _resetSinoElemt(self):
        """
        Reset the sinogram element
        """
        self._sinoElemt = {}

    def _buildAlgorithms(self):
        """
        will setup one ART algorithm per fluo sinogram
        """
        self._resetFluoSinograms()
        self.resetAngles()

        [self._buildAlgorithm(sinogram) for sinogram in self.config.sinograms]

    def __getTitle(self, header):
        """return the title of the header"""
        if 'title' in header:
            return header['title']
        if 'Title' in header:
            return header['Title']
        return None

    def __getMeanData(self, edfReader, groupFiles, indexDataToTreat, referenceName):
        """Simply create a mean sinogram from a list of files and check that
        each sinogram of the file as the same ID (referenceName)

        :param edfReader: the edfReader for the file
        :param groupFiles: the list of file definig the group
        :param indexDataToTreat: index of the sinogram in the file
        :param referenceName: The 'title' of the reference sinogram
        """
        sino = None

        for file in groupFiles:
            # check that titles are the same as the reference file. otherwise
            # through an error
            if edfReader.nframes == 1:
                if self.__getTitle(edfReader.getHeader()) != referenceName:
                    raise ValueError('incoherent title name between sinogram to compute the mean')

                if sino is None:
                    sino = edfReader.getData()
                else:
                    sino += edfReader.getData()

            else:
                if self.__getTitle(edfReader.getframe(int(indexDataToTreat)).getHeader()) != referenceName:
                    raise ValueError('incoherent title name between sinogram to compute the mean')

                if sino is None:
                    sino = edfReader.getframe(int(indexDataToTreat)).getData()
                else:
                    sino += edfReader.getframe(int(indexDataToTreat)).getData()

        return sino / len(groupFiles)

    def _buildAlgorithm(self, sinogram):
        """build the algorithm for the givem file path"""

        # read the file
        if sinogram.data is None:
            err = "sinogram %s has empty data" + sinogram.name
            raise IOError(err)

        if sinogram.fileInfo:
            datasets = self.config.getSinogramsFileDict()[sinogram.fileInfo.file_path()]
            gf, groupType = reconstrutils.tryToFindPattern(sinogram.fileInfo.file_path())
            if groupType == 'pymca':
                try:
                    self.__treatAsAPymcaGroupFiles(sinogram.fileInfo.file_path(),
                                                   datasets)
                    return
                except ValueError:
                    err = """Found incoherence in the files or in sinograms, unable
                        to associate pymca group files"""
                    _logger.info(err)
        self.__treatAsAStandardFile(sinogram)

    def __treatAsAPymcaGroupFiles(self, filePath, sinograms):
        """
        create all the algorithm from getting the mean file from the pymca
        """
        groupFiles = reconstrutils.tryToFindPattern(filePath, 'pymca')
        assert(len(groupFiles) > 0)
        if (len(groupFiles) < 2):
            err = "file can't be associated with any other pymca file"
            raise ValueError(err)

        edfReader = edfimage()
        edfReader.read(filePath)

        for sino in sinograms:
            assert(isinstance(sino, structs.FluoSino))
            name = sino.name
            self._sinoElemt[name] = sino.physElmt
            EF = sino.EF

            refName = None
            if edfReader.nframes == 1:
                refName = self.__getTitle(edfReader.getHeader())
            else:
                refName = self.__getTitle(
                    edfReader.getframe(int(sino.index)).getHeader())

                sino.data = self.__getMeanData(edfReader=edfReader,
                                               groupFiles=groupFiles,
                                               indexDataToTreat=sino.index,
                                               referenceName=refName)
            self._addFluoARTAlgorithm(sino)

    def __treatAsAStandardFile(self, sinogram):
        """
        create all the algorithm at the standard way
        """
        assert(isinstance(sinogram, structs.FluoSino))
        self._sinoElemt[sinogram.name] = sinogram.physElmt
        self._addFluoARTAlgorithm(sinogram)

    def _updateSelfAbsMat(self, fluoSinogram):
        """

        """
        assert(isinstance(fluoSinogram, structs.FluoSino))
        if self.config.reconsType == config._ReconsConfig.COMPTON_ID:
            # compton case
            assert(self.config.absMat.data is not None)
            return self.config.absMat.data
        else:
            # fluo case
            if fluoSinogram.selfAbsMat is not None:
                # use given self abs mat
                fluoSinogram.selfAbsMat.loadData()
                return fluoSinogram.selfAbsMat.data
            else:
                # create the self abs mat
                if fluoSinogram.EF is None:
                    err = "No EF given. Can't process the fluorescence reconstruction"
                    raise ValueError(err)

                assert(fluoSinogram.EF > 0.0)
                # In the case of the fluorescence
                selfAbsMatrix = np.zeros(self.config.absMat.data.shape, dtype=np.float64)
                sampleComposition = self.config.materials.matComposition.data.astype(numpy.str)
                E0 = self.config.E0
                for materialName in self.fisxMaterials:
                    EI = physicalelmts.elementInstance
                    cross_section_EF_mat = EI.getElementMassAttenuationCoefficients(materialName,
                                                                                    [fluoSinogram.EF])
                    cross_section_E0_mat = EI.getElementMassAttenuationCoefficients(materialName,
                                                                                    [E0])
                    # Note : energy is in keV in freeart AND in physx. So no need for conversion

                    sampleComposition = self._adaptMatToReduction(sampleComposition)
                    if sampleComposition.size != selfAbsMatrix.size:
                        width = min(abs(sampleComposition.shape[1] - self.config.center), self.config.center)
                        if width > 0:
                            center = self.config.center
                            _logger.info('resize sample composition matrix')
                            sampleComposition = sampleComposition[center - width:center + width, center - width:center + width]
                    condition = np.in1d(sampleComposition.ravel(), materialName[1:]).reshape(selfAbsMatrix.shape)
                    selfAbsMatrix[condition] = self.config.absMat.data[condition] * cross_section_EF_mat['total'][0] / cross_section_E0_mat['total'][0]
                # reduce the sample composition if needed (if sinogram definition
                # is reduced )
                assert(selfAbsMatrix.ndim == 3)
                selfAbsMatrix = self._adaptMatToReduction(selfAbsMatrix)
                if selfAbsMatrix.shape[1] != self._getTheoreticalSampSize():
                    s1 = selfAbsMatrix.shape[1]
                    s2 = self._getTheoreticalSampSize()
                    err = 'The self absorption matrix for (%s, %s) has invalid dimensions (%s).' \
                          'When in theory it should be (%s, %s).' % (s1, s1, s2, s2)
                    raise ValueError(err)

                assert(self.config.absMat.data.shape == selfAbsMatrix.shape)
                fluoSinogram.selfAbsMat = structs.SelfAbsMatrix(data=selfAbsMatrix)
                return selfAbsMatrix

    def _addFluoARTAlgorithm(self, fluoSinogram):
        """
        Add an ART algorithm to the set of algorithm

        :param fluoSinogram: the sinogram for which we want to build an ART
                             algorithm
        """
        assert(isinstance(fluoSinogram, structs.FluoSino))
        assert(fluoSinogram.data is not None)
        _logger.info('create the algorithm for ' + fluoSinogram.physElmt)

        selection = self.config.projections
        self.angles = self._getAngles(nbAngles=fluoSinogram.data.shape[0],
                                      selection=selection)

        fluoSinogram.data = self.adaptSinogramData(fluoSinogram.data,
                                                   isI0=False)

        # Do not do this, already managed in the freeart core process
        # if self.angles[0] != 0.0:
        #     # rotate the detector position from the origon of a certain angle
        #     self.detPos = tuple(freeart.utils.computeDetectorPosition(
        #         detPos=self.detPos,
        #         angle=self.angles[0]))
        #     self.detSetup = [(self.detPos, self.detWidth)]
        #
        #     info = 'Updating detector position for the first projection.'
        #     info += 'New position is %s' % str(self.detPos)
        #     _logger.info(info)

        # for now in compton Mode
        if not fluoSinogram.data.shape[2] == self.config.absMat.data.shape[1]:
            error = "can't run the reconstruction, width of the sinogram (%s)" \
                    "and the absorption matrix dim (%s) are differente" \
                     % (fluoSinogram.data.shape[2], self.config.absMat.data.shape[1])
            raise ValueError(error)

        selfAbsMatrix = self._updateSelfAbsMat(fluoSinogram)
        selfAbsMatrix = self._adaptMatToReduction(selfAbsMatrix)
        self.config.absMat.data = self.__checkAbsMatDim(self.config.absMat.data)
        selfAbsMatrix = self.__checkAbsMatDim(selfAbsMatrix)

        if not self.config.absMat.data.shape == selfAbsMatrix.shape:
            error = "can't run the reconstruction, shapes of the absorption " \
                    "matrix (%s )" % str(self.config.absMat.data.shape)
            error += "and shape of the self absorption matrix are not fitting" \
                     " (%s )" % str(selfAbsMatrix.shape)
            _logger.error(error)
            raise ValueError(error)

        # creating the ART algorithm
        detector = self.config.detector
        detSetup = [(detector.getPosition(), detector.width)]
        alRecons = freeart.FluoBckProjection(sinoDat=fluoSinogram.data.astype(self.precisionType),
                                             sinoAngles=self.angles,
                                             expSetUp=detSetup,
                                             absorp=self.config.absMat.data.astype(self.precisionType),
                                             selfAbsorp=selfAbsMatrix.astype(self.precisionType))

        if self.config.voxelSize:
            alRecons.setVoxelSize(self.config.voxelSize * self.config.definitionReduction)

        if self.config.oversampling:
            alRecons.setOverSampling(self.config.oversampling)

        if self.config.useAFileForI0 is False:
            alRecons.setI0(self.config.I0)

        if self.config.beamCalcMethod:
            alRecons.setRayPointCalculationMethod(self.config.beamCalcMethod)

        if self.config.outBeamCalMethod:
            alRecons.setOutgoingRayAlgorithm(self.config.outBeamCalMethod)

        if self.config.solidAngleOff:
            alRecons.turnOffSolidAngle(self.config.solidAngleOff)

        if self.config.dampingFactor:
            alRecons.setDampingFactor(self.config.dampingFactor)

        self.fluoSinogramsAlgorithms[fluoSinogram.name] = alRecons
        self.fluoSinograms[fluoSinogram.name] = fluoSinogram

        # create the interaction matrice
        self.interactionMatrice[fluoSinogram.physElmt] = self.__computeInteractionMatrice(fluoSinogram.physElmt)

    def __checkAbsMatDim(self, mat):
        if mat.ndim != 3:
            assert(mat.ndim is 2)
            mat = mat.reshape(mat.shape[0], mat.shape[1], 1)

        return mat

    def __computeInteractionMatrice(self, name):
        """Return the interaction matrice according to the materials in the
        sample and the element

        :param name: the id of the element for which we want to compute the
        interaction matrix
        """
        if self.config.reconsType == 'Compton':
            interactionMatrice = np.ones((self.config.absMat.data.shape[0],
                                          self.config.absMat.data.shape[1]),
                                          dtype=np.float64)
        else:
            interactionMatrice = np.zeros((self.config.absMat.data.shape[0],
                                           self.config.absMat.data.shape[1]),
                                          dtype=np.float64)
            materialsDict = self.config.materials.materials.data
            if materialsDict is not None:
                sampleComposition = self.config.materials.matComposition.data
                if sampleComposition.size != self.config.absMat.data.size:
                    width = min(abs(sampleComposition.shape[1] - self.config.center), self.config.center)
                    if width > 0:
                        center = self.config.center
                        _logger.info('resize sample composition matrix')
                        sampleComposition = sampleComposition[center - width:center + width, center - width:center + width]

                for materialName in materialsDict:
                    uniqMatName = reconstrutils.getFisxMatName(materialName)
                    probabilityOfEmission = FluoConfInterpreter.getProbabilityOfEmission(
                                                fisxMat=self.fisxMaterials[uniqMatName],
                                                energy=self.config.E0,
                                                element=name,
                                                fisxElements=physicalelmts.elementInstance,
                                                shell=None)
                    condition = np.in1d(sampleComposition.ravel(),
                                        materialName).reshape(interactionMatrice.shape)
                    interactionMatrice[condition] = probabilityOfEmission

            interactionMatrice.shape = (interactionMatrice.shape[0],
                                        interactionMatrice.shape[1],
                                        1)

        return interactionMatrice

    def getReconstructionAlgorithms(self):
        """
        :return: all the algorithm created for reconstructions
        """
        return self.fluoSinogramsAlgorithms

    def iterate(self, nbIteration):
        """
        Iterate on the reconstruction. Each iteration generate a backward
        projection for each angle of acquisition (in a random order).
        """
        threads = []
        nbVoxel = 0

        for index, sinogram in enumerate(self.config.sinograms):
            if index % self.limitParaRecon == 0 or (nbVoxel > self.limitVoxel):
                [thread.start() for thread in threads]
                [thread.join() for thread in threads]
                threads = []

            assert(not self.fluoSinogramsAlgorithms[sinogram.name] is None)
            threads.append(Thread(target=self.fluoSinogramsAlgorithms[sinogram.name].iterate,
                                  args=(nbIteration,)))
            nbVoxel = self.fluoSinograms[sinogram.name].data.shape[1] * self.fluoSinograms[sinogram.name].data.shape[1]

        [thread.start() for thread in threads]
        [thread.join() for thread in threads]

    def setRandSeedToZero(self, b):
        for sinogramName in self.fluoSinogramsAlgorithms:
            assert(self.fluoSinogramsAlgorithms[sinogramName] is not None)
            self.fluoSinogramsAlgorithms[sinogramName].setRandSeedToZero(b)

    def getDensityPhantoms(self):
        """
        Return the phantom taking into account the probability of generation
        of the physical element
        """
        res = {}
        for element in self.fluoSinogramsAlgorithms:
            ph = np.zeros(self.interactionMatrice[element].shape)
            if self.interactionMatrice[element] is not None:
                probInter = self.fluoSinogramsAlgorithms[element].getPhantom()[self.interactionMatrice[element] > 0.0] / \
                                                            self.interactionMatrice[element][self.interactionMatrice[element] > 0.0]
                ph[self.interactionMatrice[element] > 0.0] = probInter
            res[element] = ph

        return res

    def getInteractionPhantoms(self):
        return self.interactionMatrice



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2016 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/
# __authors__ = ["H. Payno"]
# __license__ = "MIT"
# __date__ = "01/09/2016"

# import unittest
# import os
# import tempfile
# import sys
# if sys.version < '3.0':
#     import ConfigParser as configparser
# else:
#     import configparser

# from freeart.interpreter import GlobalConfigInterpreter
# import numpy
# from freeart.utils import reconstrutils, testutils, physicalelmts, genph, raypointsmethod, outgoingrayalgorithm
# import freeart
# from freeart.interpreter.config import Config
# from freeart.unitsystem import metricsystem

# from fisx import Material

# import numpy as np
# import logging
# _logger = logging.getLogger("unitests")

# class testConfigInterpreterFluo(unittest.TestCase):
#     """Test that we are able to generate a valid fluorescence reconstruction using a cfg file and the interpreter"""

#     def setUp(self):
#         materialSteel = {'Comment':"No comment",
#                  'CompoundList':['Cr', 'Fe', 'Ni'],
#                  'CompoundFraction':[18.37, 69.28, 12.35],
#                  'Density':1.0,
#                  'Thickness':1.0}

#         materialTeflon = {'Comment':"No comment",
#                  'CompoundList':['C1', 'F1'],
#                  'CompoundFraction':[0.240183, 0.759817],
#                  'Density':1.0,
#                  'Thickness':1.0}

#         materials = {"Steel":materialSteel, "Teflon":materialTeflon}
#         sheppLoganPartID = {"Steel":3, "Teflon":2}
#         densities = {"Steel":0.5, "Teflon":0.2}
#         EF2 = {"Steel":1.5e5, "Teflon":2.4e4}

#         self.experimentationTwoMat = testutils.FluorescenceExperimentationExWithFisx(materials, densities, 
#             oversampling=20, dampingFactor=0.02, E0=1.0e6, EF=EF2, nbAngles=400, sinoElmts=['Cl', 'Fe', 'Ni'])
#         self.experimentationTwoMat.setUp()
#         self.experimentationTwoMat.voxelSize = 3.0*metricsystem.cm

#     def testWithFisx(self):
#         exp = self.experimentationTwoMat
#         interpreter = GlobalConfigInterpreter(exp.cfg_fluo_file)
#         # for unit tests, reset the random generator
#         interpreter.setRandSeedToZero(True)
#         interpreter.iterate(3)

#         algos = interpreter.getReconstructionAlgorithms()
#         # check parameters from the first algorithm
#         for material in algos :
#             reconstructed_phantom = algos[material].getPhantom()
#             assert(material in exp.phantoms)
#             numpy.testing.assert_allclose(reconstructed_phantom, exp.phantoms[material], rtol=0.3, atol=5e-2)  

#     def tearDown(self):
#         self.experimentationSteel.tearDown()
#         self.experimentationTwoMat.tearDown()
#         del self.experimentationSteel
#         del self.experimentationTwoMat


# def suite():
#     test_suite = unittest.TestSuite()
#     test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testWithFisx))
    
#     return test_suite

# if __name__ == '__main__':
#     unittest.main(defaultTest="suite")        


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

import unittest
import numpy as np

import freeart
from freeart.unitsystem import metricsystem
from freeart.utils import reconstrutils, raypointsmethod, outgoingrayalgorithm

import logging
_logger = logging.getLogger("unitests")

class testFluoSubdivisionSampling(unittest.TestCase):
    """The  unit test to check if the option matriceSubdivision for computing the values of the outgoing attenuation is correct.
    Note : previously we had tests for incoming or outgoing beam attenaution effect using the option withInterpolation.
    This is not very relevant here because the interpolation will get the value fron the close voxels for the density (phantom), 
    the incoming beam attenuation (absMat) AND the outgoing beam attenaution. So this complexifiate the theoretical value
    we are supposed to get. Or to simplify we have to take large phantom and this slow down the results.
    """

    phDimX = 3
    phDimY = 3
    phDimZ = 1

    detectorRadius = 10.0 * metricsystem.cm
    detectorDistance = 1000.0 * metricsystem.cm

    def test_outgoingRayCalculation(self):           
        """Test the attenuation for the outgoing beam hitting n lines of material.
        Material attenuation of the incoming beam is zero.
        Material attenuation of the outgoing beam is outgoingAbsorptionFactor 
        To make sure results are ok we are testing different detector position, 
        which are directly defining the angle of the outgoing beam.
        """

        # The outgoing beam attenuation to consider for this test case
        outgoingAbsorptionFactor = 1.0  # g/cm2
        interactionProba = 1.0

        detAngle = np.pi/4.0
        oversamplings = [100]
        subdivisionFactors = [10]

        methodGBeams = raypointsmethod.withoutInterpolation

        detPos = (np.sin(detAngle)*self.detectorDistance, np.cos(detAngle)*self.detectorDistance, 0)
        detSetup = [(detPos, self.detectorRadius)]

        ### RUN the FREE ART EXPERIMENTATION
        # create the simple phantom
        phantom    = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)
        absMat     = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)
        selfAbsMat = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=np.float64)

        phantom[1, 1, 0]    = interactionProba
        selfAbsMat[1, 1, 0] = outgoingAbsorptionFactor

        for oversampling in oversamplings : 
            # The value here for an angle of 45 degree should get close to 1/2 with a constant solid angle of 1 
            # will cover the full voxel. So beams hitting the detector should cover the entire voxel and go throught a mean of half the voxel
            # so the result should approach np.exp(-1.0*0.5)
            _logger.info("\n\n OVERSAMPLING = " + str(oversampling))
            _logger.info("\n -- RAW --  ")
                
            optionOutgoing = outgoingrayalgorithm.rawApproximation
            sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
                phantom, absMat, selfAbsMat, 1, detSetup, oversampling, methodGBeams, optionOutgoing, turnOffSolidAngle = True );
            theoreticalResult = np.exp(-np.sqrt(2)*0.5)
            np.testing.assert_allclose(sinogram[0][0], [0.0, theoreticalResult, 0.0], rtol=0.1 )

            _logger.info("\n -- CREATE ONE RAY PER SAMPLE POINT --  ")
    
            # this method gave the true attenaution off the outgoing ray. But we still have to sample enought point to get the true result.
            optionOutgoing = outgoingrayalgorithm.createOneRayPerSamplePoint
            sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
                phantom, absMat, selfAbsMat, 1, detSetup, oversampling, methodGBeams, optionOutgoing, turnOffSolidAngle = True );

            theoreticalResult = np.exp(-1*0.5)
            np.testing.assert_allclose(sinogram[0][0], [0.0, theoreticalResult, 0.0], rtol=0.1 )

            _logger.info(" -- MATRICE SUBDIVISION -- ")

            for subdivisionFactor in subdivisionFactors : 
                _logger.info("subdivision factor : " + str(subdivisionFactor))
    
                optionOutgoing = outgoingrayalgorithm.matriceSubdivision
        
                sinogram, angles = reconstrutils.makeFreeARTFluoSinogram    (
                    phantom, absMat, selfAbsMat, 1, detSetup, oversampling, methodGBeams, optionOutgoing, subdivisionValue=subdivisionFactor, turnOffSolidAngle=True );

                theoreticalResult = np.exp(-(interactionProba*outgoingAbsorptionFactor*0.5))
                _logger.info("res : " + str(sinogram[0][0][1]))
                _logger.info("theoretical res : " + str(theoreticalResult))

                np.testing.assert_allclose(sinogram[0][0], [0.0, theoreticalResult, 0.0], rtol=0.1 )
                
def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testFluoSubdivisionSampling))
    
    return test_suite

if __name__ == '__main__':
    unittest.main(defaultTest="suite")


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

import unittest
import numpy as np

import freeart
from freeart.utils import unittestpercentage, raypointsmethod, outgoingrayalgorithm, reconstrutils, testutils
from freeart.unitsystem import metricsystem

class testUnits(unittestpercentage.testPercentage):
    """Here we want to test the incoming beam attenuation in a stacks of voxels."""

    origin          = (0.0, 0.0)
    detectorRadius  = 10.0 *metricsystem.cm
    detectorArea    = detectorRadius * 2.0
    detectorWidth   = detectorRadius*2.0
    detectorDistance = 1000 *metricsystem.cm

    def testCompareMillimeterAndCentimeter(self):
        """
        A simple test to make sure absortpion across a voxel of 1cm is close to the absorption of 10 voxels of 1mm
        if the absorption in those voxels (in g/cm2) is the same 
        """
        incomingAbsorption = 1.0
        interactionProba = 1.0
        outgoingAbsorptionFactor = 0.0 # g/cm2
        phDimX = 31 # voxel
        phDimY = 31 # voxel
        phDimZ = 1  # voxel

        # to avoid bord effect it is better to avoid interpolation 
        methodGBeams = raypointsmethod.withInterpolation
        # and set sampling to 1
        oversampling = 4
        optionOutgoing = outgoingrayalgorithm.rawApproximation

        detAngle = 0
        detPos = (np.sin(detAngle)*self.detectorDistance, np.cos(detAngle)*self.detectorDistance, 0.0)
        detSetup = [(detPos, self.detectorRadius)]
        
        ### RUN the FREE ART EXPERIMENTATION
        phantom_cm = np.zeros((phDimX, phDimY, phDimZ), dtype=np.float64)
        phantom_cm[15, 15, 0] = interactionProba
        absMat_cm = np.zeros((phDimX, phDimY, phDimZ), dtype=np.float64)
        absMat_cm[15, 15, 0] = incomingAbsorption
        selfAbsMat_cm = np.zeros((phDimX, phDimY, phDimZ), dtype=np.float64)
        selfAbsMat_cm[15, 15, 0] = outgoingAbsorptionFactor

        phantom_mm = np.zeros((phDimX, phDimY, phDimZ), dtype=np.float64)
        phantom_mm[8:18, 8:18, 0] = interactionProba
        absMat_mm = np.zeros((phDimX, phDimY, phDimZ), dtype=np.float64)
        absMat_mm[8:18, 8:18, 0] = incomingAbsorption
        selfAbsMat_mm = np.zeros((phDimX, phDimY, phDimZ), dtype=np.float64)
        selfAbsMat_mm[8:18, 8:18, 0] = outgoingAbsorptionFactor

        sinogram_cm, angles = reconstrutils.makeFreeARTFluoSinogram(
            phantom_cm, absMat_cm, selfAbsMat_cm, 1, detSetup, oversampling, methodGBeams, optionOutgoing, 
            voxelSize=1.0*metricsystem.cm, turnOffSolidAngle=True );

        sinogram_mm, angles = reconstrutils.makeFreeARTFluoSinogram(
            phantom_mm, absMat_mm, selfAbsMat_mm, 1, detSetup, oversampling, methodGBeams, optionOutgoing, 
            voxelSize=1.0*metricsystem.mm, turnOffSolidAngle=True );

        # sinogram_mm return It ot in the case of voxel set to 1 mm then as the absortpion is computing 
        # on the voxel surface the ray will be less absorbed.
        self.assertAlmostEqual( sinogram_mm[0][0][15], sinogram_cm[0][0][15], 1)

    def testFluoRecons(self):
        """
        Simple test on having a fluo reconstruction with a voxel size set to 1.2 mm
        """
        oversampling = 12
        nbIterations = 3
        voxelSize = 1.2*metricsystem.millimeter

        detPos = (1000*metricsystem.cm, 0, 0)
        detSetup = [(detPos, self.detectorRadius)]
        sheppLogan = testutils.SheppLoganReconstruction()

        sino, angles, absMat, selfAbsMat, initialPhantom = sheppLogan.produce_SheppLogan_sinogram(32, _oversampling=oversampling, 
            _anglesNb=400, _detSetup=detSetup, _voxelSize = voxelSize )

        reconstructedPhantom = sheppLogan.make_reconstruction(_sinogram=sino, _angles=angles, _absMat=absMat, 
            _selfAbsMat=selfAbsMat, _detSetup=detSetup, _nbIter=nbIterations, _oversampling=oversampling, 
            _voxelSize = voxelSize, _dampingFactor=0.05)

        np.testing.assert_allclose(reconstructedPhantom, initialPhantom, rtol=0.15, atol=1e-3)

def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testUnits))
    return test_suite

if __name__ == '__main__' and __package__ is None :
    unittest.main(defaultTest="suite")


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

"""
Outgoing beam calculation method
This define the way the OUTGOING ray values are computed
This apply only for the fluorescence mode

rawApproximation : in this case we are making the evaluation of the outgoing ray the faster as possible
                   by computing for each rotation the mean value of some outgoing rays and taking it as the value
                   of the voxel

createOneRayPerSamplePoint : in this version we are creating the real outgoing ray for each sample point on the incoming ray
                    and computing the 'real' outgoing absorption of the ray by sampling voxels           
matriceSubdivision : this is the raw approximation case but by subdividing the voxels for the outgoing rays
                     we are reducing the noise of the "raw" approximation.
"""

rawApproximation = 0
createOneRayPerSamplePoint = 1
matriceSubdivision = 2




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tool to provide Shepp-Logan phantoms."""
__authors__ = ["N. Vigano"]
__license__ = "MIT"
__date__ = "01/09/2014"

import sys
import numpy

class PhantomGenerator(object):
    '''
    Class for generating different Phantoms
    '''
    class Ellipsoid:
        def __init__(self, a, b, c, x0, y0, z0, alpha, mu):
            self.a = a; self.b = b; self.c = c
            self.x0 = x0; self.y0 = y0; self.z0 = z0
            self.alpha = alpha * numpy.pi / 180.0; self.mu = mu
            self.cosAlpha = numpy.cos(self.alpha)
            self.sinAlpha = numpy.sin(self.alpha)
    
    sheppLogan = [
        #         a       b      c       x0     y0      z0     alpha  mu
        Ellipsoid(0.69,   0.92,  0.90,   0.0,   0.0,    0.0,    0.0,  0.10),
        Ellipsoid(0.6624, 0.874, 0.88,   0.0,  -0.02,   0.0,    0.0, -0.08),
        Ellipsoid(0.11,   0.31,  0.21,   0.22, -0.0,    0.0,  -18.0, -0.02),
        Ellipsoid(0.16,   0.41,  0.22,  -0.22,  0.0,   -0.25,  18.0, -0.02),
        Ellipsoid(0.21,   0.25,  0.35,   0.0,   0.35,  -0.25,   0.0,  0.03),
        Ellipsoid(0.046,  0.046, 0.046,  0.0,   0.10,  -0.25,   0.0,  0.01),
        Ellipsoid(0.046,  0.046, 0.02,   0.0,  -0.10,  -0.25,   0.0,  0.01),
        Ellipsoid(0.046,  0.023, 0.02,  -0.08, -0.605, -0.25,   0.0,  0.01),
        Ellipsoid(0.023,  0.023, 0.10,   0.0,  -0.605, -0.25,   0.0,  0.01),
        Ellipsoid(0.023,  0.046, 0.10,   0.06, -0.605, -0.25,   0.0,  0.01)
    ]

    metalPhantom = [
        #         a       b      c       x0       y0      z0     alpha  mu
        Ellipsoid(0.95,   0.95,  0.95,   0.0,     0.0,    0.0,    0.0,  0.02),
        Ellipsoid(0.15,   0.15,  0.15,   0.0,     0.6,    0.0,    0.0,  0.08),
        Ellipsoid(0.10,   0.10,  0.10,   0.4242,  0.4242, 0.0,    0.0,  0.08),
        Ellipsoid(0.05,   0.05,  0.05,   0.6,     0.0,    0.0,    0.0,  0.08),
        Ellipsoid(0.02,   0.02,  0.02,   0.4242, -0.4242, 0.0,    0.0,  0.08),
        Ellipsoid(0.15,   0.15,  0.15,   0.0,    -0.6,    0.0,    0.0,  0.88),
        Ellipsoid(0.10,   0.10,  0.10,  -0.4242, -0.4242, 0.0,    0.0,  0.88),
        Ellipsoid(0.05,   0.05,  0.05,  -0.6,     0.0,    0.0,    0.0,  0.88),
        Ellipsoid(0.02,   0.02,  0.02,  -0.4242,  0.4242, 0.0,    0.0,  0.88)
    ]

    def __init__(self):
        pass

    def get3DPhantomSheppLogan(self, n):
        """
        :param n: The width (and height) of the phantom to generate 
        :return: A numpy array of dimension n*n fit with the sheppLogan phantom
        """
        volume = numpy.ndarray(shape=(n, n, n))
        volume.fill(0.)
        
        count = 0
        for ell in PhantomGenerator.sheppLogan:
            count = count+1
            # print("Ellisse n: %d con mu: %f" % (count, ell.mu))
            squareZ = self._getSquareZ(n, ell)
            squareZ.shape = (1,n)
            for x in range(n):
                sumSquareXandY = self._getSquareXandYsum(n, x, ell)
                sumSquareXandY.shape = (n,1)
                sumSquareXYZ = numpy.add(squareZ, sumSquareXandY)
                indices = sumSquareXYZ <= 1
                volume[x,indices] = ell.mu
        indices = numpy.abs(volume) > 0
        volume[indices] = numpy.multiply(volume[indices] + 0.1, 5)
        return volume

    def get2DPhantomSheppLogan(self, n, ellipsoidID=None):
        """
        :param n: The width (and height) of the phantom to generate 
        :param ellipsoidID: The Id of the ellipsoid to pick. If None will produce all the ellipsoid
        """
        assert(ellipsoidID is None or (ellipsoidID >=0 and ellipsoidID < len(PhantomGenerator.sheppLogan)))
        if ellipsoidID is None :
            area = self._get2DPhantom(n, PhantomGenerator.sheppLogan)
        else :
            area = self._get2DPhantom(n, [PhantomGenerator.sheppLogan[ellipsoidID]])

        indices = numpy.abs(area) > 0
        area[indices] = numpy.multiply(area[indices] + 0.1, 5)
        return area / 100.0

    def get2DPhantomMetal(self, n, ellipsoidID=None):
        """
        :param n: The width (and height) of the phantom to generate 
        :param ellipsoidID: The Id of the ellipsoid to pick. If None will produce all the ellipsoid
        """        
        assert(ellipsoidID is None or (ellipsoidID >=0 and ellipsoidID < len(PhantomGenerator.metalPhantom)))
        if ellipsoidID is None :
            area = self._get2DPhantom(n, PhantomGenerator.metalPhantom)
        else :
            area = self._get2DPhantom(n, [PhantomGenerator.metalPhantom[ellipsoidID]])
        area = numpy.multiply(area, 10)
        return area / 100.0

    def _get2DPhantom(self, n, phantomSpec):
        area = numpy.ndarray(shape=(n, n))
        area.fill(0.)

        count = 0
        for ell in phantomSpec:
            count = count+1
            for x in range(n):
                sumSquareXandY = self._getSquareXandYsum(n, x, ell)
                indices = sumSquareXandY <= 1
                area[indices, x] = ell.mu
        return area

    def _getSquareXandYsum(self, n, x, ell):
        div = lambda x, y: numpy.divide(x, y)
        mul = lambda x, y: numpy.multiply(x, y)
        sub = lambda x, y: numpy.subtract(x, y)
        add = lambda x, y: numpy.add(x, y)
        pow2 = lambda x: numpy.power(x, 2)

        supportX1 = numpy.ndarray(shape=(n, ))
        supportX2 = numpy.ndarray(shape=(n, ))
        support_consts = numpy.ndarray(shape=(n, ))

        xScaled = float(2*x-n)/float(n)
        xCos =  xScaled*ell.cosAlpha
        xSin = -xScaled*ell.sinAlpha
        supportX1.fill(xCos)
        supportX2.fill(xSin)

        supportY1 = numpy.arange(n)
        support_consts.fill(2.)
        supportY1 = mul(support_consts, supportY1)
        support_consts.fill(n)
        supportY1 = sub(supportY1, support_consts)
        support_consts.fill(n)
        supportY1 = div(supportY1, support_consts)
        supportY2 = numpy.array(supportY1)

        support_consts.fill(ell.sinAlpha)
        supportY1 = add(supportX1, mul(supportY1, support_consts))
        support_consts.fill(ell.cosAlpha)
        supportY2 = add(supportX2, mul(supportY2, support_consts))

        support_consts.fill(ell.x0)
        supportY1 = sub(supportY1, support_consts)
        support_consts.fill(ell.y0)
        supportY2 = sub(supportY2, support_consts)

        support_consts.fill(ell.a)
        supportY1 = pow2(div(supportY1, support_consts))
        support_consts.fill(ell.b)
        supportY2 = pow2(div(supportY2, support_consts))

        return add(supportY1, supportY2)

    def _getSquareZ(self, n, ell):
        div = lambda x, y: numpy.divide(x, y)
        mul = lambda x, y: numpy.multiply(x, y)
        sub = lambda x, y: numpy.subtract(x, y)
        pow2 = lambda x: numpy.power(x, 2)

        supportZ1 = numpy.arange(n)
        support_consts = numpy.ndarray(shape=(n, ))
        support_consts.fill(2.)
        supportZ1 = mul(support_consts, supportZ1)
        support_consts.fill(n)
        supportZ1 = sub(supportZ1, support_consts)
        support_consts.fill(n)
        supportZ1 = div(supportZ1, support_consts)

        support_consts.fill(ell.z0)
        supportZ1 = sub(supportZ1, ell.z0)

        support_consts.fill(ell.c)
        return pow2(div(supportZ1, support_consts))

    def print2DPhantomToPlainTxt(self, ph, file="output.txt"):
        f = open(file,"w")
        shp = ph.shape
        f.write("%d\n" % shp[0])
        f.write("%d\n" % shp[1])
        f.write("1\n")
        for y in range(shp[1] - 1,0,-1):
            for x in range(shp[0]):
                f.write("%1.7e " % ph[y,x])
                if (x >= (shp[0]-1)):
                    f.write("\n")
        f.close()


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("genph usage: genph <Phantom file name>")
        sys.exit(-1)

    phgen = PhantomGenerator()
    phantom = phgen.get2DPhantomSheppLogan(256)
    phgen.print2DPhantomToPlainTxt(phantom,sys.argv[1])




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Contains the python representation of the reconstruction parameters
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"


from freeart.unitsystem import metricsystem
from silx.io import url
from silx.io import configdict
import numpy
import time
try:
    from freeart import version
except:
    version = ''
from . import structs, fileInfo
import os


class _ReconsConfig(object):

    TX_ID = "Transmission"
    COMPTON_ID = "Compton"
    FLUO_ID = "Fluorescence"

    RECONS_TYPES = [TX_ID, COMPTON_ID, FLUO_ID]

    PRECISION_TO_TYPE = {
        'simple': numpy.float32,
        'double': numpy.float64
    }

    reconsParams = None
    """
    The reconstruction parameters (oversampling...) to launch once configured
    """

    def __init__(self, reconsType=None, minAngle=None, maxAngle=None,
                 projections=None, defReduction=1, projNumReduction=1,
                 oversampling=1, solidAngleOff=False,
                 includeLastProjection=True, I0=None, beamCalcMethod=0,
                 centerOfRotation=-1, acquiInverted=False, precision='double'):
        """

        :param reconsType:
        :param centerOfRotation: if -1 then will take the middle
        :param precision: calcul precision, can be simple or double.
        """
        if reconsType not in _ReconsConfig.RECONS_TYPES:
            raise ValueError('reconsType not recognized')
        if minAngle is not None and not (0 <= minAngle <= 2.*numpy.pi):
            raise ValueError('min angle not in [0, 2pi]')
        if minAngle is not None and not (0 <= maxAngle <= 2.*numpy.pi):
            raise ValueError('max angle not in [0, 2pi]')
        if (minAngle is not None or maxAngle is not None) and not minAngle <= maxAngle:
            raise ValueError('False condition: minAngle <= maxAngle')

        self.precision = precision
        assert self.precision in self.PRECISION_TO_TYPE
        self.reconsType = reconsType
        self.center = centerOfRotation
        self.projections = projections
        """The projections to be consider for the reconstruction"""
        self.definitionReduction = defReduction
        """
        Should we take a pixel each definition_reduction for the reconstruction
        """
        self.projectionNumberReduction = projNumReduction
        """
        Should we take one projection each projection_number_reduction for the
         reconstruction.
        """
        self.voxelSize = metricsystem.cm
        """voxel size"""
        self.oversampling = oversampling
        """Oversampling value"""
        self.solidAngleOff = solidAngleOff
        """Should we always set the the value of the solid angle to 1"""
        self.minAngle = minAngle
        self.maxAngle = maxAngle
        self.includeLastProjection = includeLastProjection
        self.setI0(I0)
        self.beamCalcMethod = beamCalcMethod
        self.acquiInverted = acquiInverted

    def setI0(self, I0):
        if type(I0) is str or (hasattr(I0, 'dtype') and numpy.issubdtype(I0.dtype,
                                                                   numpy.character)):
            self._I0 = structs.I0Sinogram()
            _url = url.DataUrl(path=I0)
            if _url.file_path():
                file_extension = _url.file_path().split('.')[-1]
            else:
                file_extension = ''
            if file_extension.lower() == 'edf':
                self._I0.fileInfo = fileInfo.MatrixFileInfo(file_path=_url.file_path(),
                                                            data_slice=_url.data_slice())
                assert self._I0.fileInfo.is_valid()
            elif file_extension.lower() in ('h5', 'hdf5', 'nx', 'nxs'):
                self._I0.fileInfo = fileInfo.MatrixFileInfo(file_path=_url.file_path(),
                                                            data_path=_url.data_path())
                assert self._I0.fileInfo.is_valid()
            self.useAFileForI0 = True
        elif isinstance(I0, structs.I0Sinogram):
            self._I0 = I0
            self.useAFileForI0 = True
        else:
            self._I0 = 1.0 if I0 is None else I0
            self.useAFileForI0 = False

    def getI0(self):
        return self._I0

    I0 = property(getI0, setI0)

    def toDict(self):
        """Convert the configuration to a silx ConfigDict"""
        dic = configdict.ConfigDict()
        dic['general_settings'] = self._getGeneralSettings()
        dic['normalization'] = self._getNormalization()
        dic['reconstruction_properties'] = self._getReconsProp()
        dic['reduction_data'] = self._getReductionDataInfo()
        dic['projection_information'] = self._getProjInfo()
        return dic

    def _getGeneralSettings(self):
        return {
            'reconstruction_type': self.reconsType,
            'date ': str(time.strftime("%d/%m/%Y")),
            'freeart_version': version,
            'precision': self.precision
        }

    def _getNormalization(self):
        if isinstance(self.I0, structs.DataStored):
            if self.I0.fileInfo is None:
                i0_out = ''
            else:
                i0_out = self.I0.fileInfo.path()
        else:
            i0_out = self.I0
        return {
            'rotation_center': self.center,
            'normalizationi0fromafile': self.useAFileForI0,
            'i0': i0_out,
        }

    def _getReconsProp(self):
        return {
            'voxel_size': self.voxelSize,
            'oversampling': self.oversampling,
            'bean_calculation_method': self.beamCalcMethod,
            'solid_angle_is_off': self.solidAngleOff,
            'include_last_angle': self.includeLastProjection
        }

    def _getReductionDataInfo(self):
        return {
            'definition_reducted_by': self.definitionReduction,
            'projection_number_reducted_by': self.projectionNumberReduction,
        }

    def _getProjInfo(self):
        return {
            'min_angle': self.minAngle,
            'max_angle': self.maxAngle,
            'projections_sel': self.projections or ':',
            'acqui_inv': self.acquiInverted
        }

    def _fromDict(self, dict):
        """set the configuration fron a silx ConfigDict"""
        self._setGenSettingFrmDict(dict)
        self._setNormalizationFrmDict(dict)
        self._setReconsPropFrmDict(dict)
        self._setReducDataFrmDict(dict)
        self._setProjInfoFrmDict(dict)
        return self

    def _setGenSettingFrmDict(self, dict):
        assert('general_settings' in dict)
        assert('reconstruction_type' in dict['general_settings'])
        self.reconsType = dict['general_settings']['reconstruction_type']
        if type(self.reconsType) is not str:
            self.reconsType = str(self.reconsType.tostring().decode())
        assert('precision' in dict['general_settings'])
        self.precision = dict['general_settings']['precision']
        if type(self.precision) is not str:
            self.precision = str(self.precision.tostring().decode())

    def _setNormalizationFrmDict(self, dict):
        assert('normalization' in dict)
        info = dict['normalization']
        assert('rotation_center' in info)
        assert('i0' in info)
        self.center = info['rotation_center']
        i0 = info['i0']
        if hasattr(i0, 'dtype') and not numpy.issubdtype(i0.dtype, numpy.number):
            i0 = str(i0.tostring().decode())
        self.setI0(i0)

    def _setReconsPropFrmDict(self, dict):
        assert('reconstruction_properties' in dict)
        info = dict['reconstruction_properties']
        assert('voxel_size' in info)
        assert('oversampling' in info)
        assert('bean_calculation_method' in info)
        assert('solid_angle_is_off' in info)
        assert('include_last_angle' in info)
        self.voxelSize = float(info['voxel_size'])
        self.oversampling = int(info['oversampling'])
        self.beamCalcMethod = int(info['bean_calculation_method'])
        self.solidAngleOff = bool(info['solid_angle_is_off'])
        self.includeLastProjection = bool(info['include_last_angle'])

    def _setReducDataFrmDict(self, dict):
        assert('reduction_data' in dict)
        info = dict['reduction_data']
        assert('definition_reducted_by' in info)
        assert('projection_number_reducted_by' in info)
        self.definitionReduction = info['definition_reducted_by']
        self.projectionNumberReduction = info['projection_number_reducted_by']

    def _setProjInfoFrmDict(self, dict):
        assert('projection_information' in dict)
        info = dict['projection_information']
        assert('min_angle' in info)
        assert('max_angle' in info)
        assert('projections_sel' in info)
        assert('acqui_inv' in info)
        self.minAngle = info['min_angle']
        self.maxAngle = info['max_angle']
        self.projections = info['projections_sel']
        if not type(self.projections) is str:
            self.projections = str(self.projections.tostring().decode())
        self.acquiInverted = bool(info['acqui_inv'])

    def __eq__(self, other):
        projectionsAreEqual = self.projections == other.projections or \
            self.projections is None and other.projections == ':' or \
            self.projections == ':' and other.projections is None

        return self.reconsType == other.reconsType and \
               self.maxAngle == other.maxAngle and \
               self.minAngle == other.minAngle and \
               self.I0 == other.I0 and \
               self.center == other.center and \
               self.includeLastProjection == other.includeLastProjection and \
               self.useAFileForI0 == other.useAFileForI0 and \
               self.solidAngleOff == other.solidAngleOff and \
               self.beamCalcMethod == other.beamCalcMethod and \
               self.voxelSize == other.voxelSize and \
               self.definitionReduction == other.definitionReduction and \
               projectionsAreEqual and \
               self.projectionNumberReduction == other.projectionNumberReduction and \
               self.acquiInverted == other.acquiInverted and \
               self.precision == other.precision


    def __str__(self):
        res = 'reconsType is %s \n' % self.reconsType
        res = res + 'oversampling is %s \n' % self.oversampling
        res = res + 'minAngle is %s \n' % self.minAngle
        res = res + 'maxAngle is %s \n' % self.maxAngle
        res = res + 'I0 is %s \n' % self.I0
        res = res + 'center is %s \n' % self.center
        res = res + 'includeLastProjection is %s \n' % self.includeLastProjection
        res = res + 'useAFileForI0 is %s \n' % self.useAFileForI0
        res = res + 'solidAngleOff is %s \n' % self.solidAngleOff
        res = res + 'beamCalcMethod is %s \n' % self.beamCalcMethod
        res = res + 'voxelSize is %s \n' % self.voxelSize
        res = res + 'definitionReduction is %s \n' % self.definitionReduction
        res = res + 'projections is %s \n' % self.projections
        res = res + 'acquiInverted is %s \n' % self.acquiInverted
        res = res + 'projectionNumberReduction is %s \n' % self.projectionNumberReduction
        return res

    def setFileInfoToH5File(self, refFile, save=False):
        """
        Change all the fileInfo to point ot the given h5 file

        :param refFile: the h5 file to store information
        :param save: if true once the file info updated, save the dataset
        """
        assert fileInfo
        fn, file_extension = os.path.splitext(refFile)
        assert file_extension in ('.hdf5', '.hdf', '.h5')
        if self.I0 is not None and self.useAFileForI0 is True:
            self.I0.loadData()
            self.I0.fileInfo = fileInfo.MatrixFileInfo(file_path=refFile,
                                                       data_path=structs.I0Sinogram.I0_DATASET)
            if save is True:
                self.I0.save()


class _ItReconsConfig(_ReconsConfig):
    """Define parameters needed for an iterative reconstruction

    :param dampingFactor: if the dampingFactor is None then will be deduced
                          automatically
    """
    def __init__(self, dampingFactor=None, *var, **kw):
        _ReconsConfig.__init__(self, *var, **kw)
        self.dampingFactor = dampingFactor

    def _getReconsProp(self):
        dict = _ReconsConfig._getReconsProp(self)
        dict['relaxation_factor'] = self.dampingFactor
        return dict

    def _setReconsPropFrmDict(self, dict):
        _ReconsConfig._setReconsPropFrmDict(self, dict)
        assert('relaxation_factor' in dict['reconstruction_properties'])
        self.dampingFactor = dict['reconstruction_properties']['relaxation_factor']

    def __eq__(self, other):
        return _ReconsConfig.__eq__(self, other) and self.dampingFactor == other.dampingFactor


class _SimpleSinoConfig(_ItReconsConfig):
    def __init__(self, reconsType, sinograms=None, computeLog=True, *var, **kw):
        _ItReconsConfig.__init__(self, *var, reconsType=reconsType, **kw)
        self.sinograms = []
        """list of sinograms to be reconstructed"""
        if sinograms:
            for sinogram in sinograms:
                assert isinstance(sinogram, self._defaultSinoType())
                self.addSinogram(sinogram)

    def addSinogram(self, sinogram):
        if sinogram is None:
            return
        elif isinstance(sinogram, self._defaultSinoType()):
            _sinogram = sinogram
        # case of h5 file
        elif type(sinogram) is str or (hasattr(sinogram, 'dtype') and
                                           numpy.issubdtype(sinogram.dtype, numpy.character)):
            _sinogram = self._defaultSinoType()()
            # Raw check for hdf5 file
            if '::' in str(sinogram):
                _sinogram.fileInfo = fileInfo.MatrixFileInfo(file_path=str(sinogram))
            else:
                _sinogram.fileInfo = fileInfo.MatrixFileInfo(file_path=str(sinogram))
        else:
            _sinogram = self._defaultSinoType()()
            _sinogram.data = sinogram

        assert isinstance(_sinogram, self._defaultSinoType())
        self.sinograms.append(_sinogram)

    def _defaultSinoType(self):
        return structs.Sinogram

    def setFileInfoToH5File(self, refFile, save=False):
        """
        Change all the fileInfo to point ot the given h5 file

        :param refFile: the h5 file to store information
        :param save: if true once the file info updated, save the dataset
        """
        _ItReconsConfig.setFileInfoToH5File(self, refFile, save)
        for _sinogram in self.sinograms:
            had_data = self.data is not None
            _sinogram.loadData()
            has_data = self.data is not None
            # assert had_data == has_data or had_data is False
            _sinogram.fileInfo = fileInfo.MatrixFileInfo(file_path=refFile,
                                                         data_path=_sinogram.h5defaultPath)
            has_data = self.data is not None
            # assert had_data == has_data or had_data is False

            if save is True:
                _sinogram.save()
                has_data = self.data is not None
                # assert had_data == has_data or had_data is False

    def toDict(self):
        dic = _ItReconsConfig.toDict(self)
        dic[self._getDataSourceHeader()] = self._getDataSourceInfo()
        self._sinogramsToDict(dic)
        return dic

    def _sinogramsToDict(self, dic):
        sinogramsDict = self.getSinogramsFileDict()
        for iSinoFile, sinoFile in enumerate(sinogramsDict):
            index = 0
            section = 'sino_file_' + str(iSinoFile)
            dic[section] = {}
            for sinogram in sinogramsDict[sinoFile]:
                assert isinstance(sinogram, self._defaultSinoType())
                if sinogram.fileInfo is not None:
                    url = sinogram.fileInfo.path()
                    dic[section]['sino_url_' + str(index)] = url
                    dataNameID = 'data_name_' + str(index)
                    dic[section][dataNameID] = sinogram.name
                    index += 1
        return dic

    def _getDataSourceHeader(self):
        return 'data_source_tx'

    def _getDataSourceInfo(self):
        return {}

    def _fromDict(self, dict):
        _ItReconsConfig._fromDict(self, dict)
        self._sinogramsFromDict(dict)
        return self

    def _sinogramsFromDict(self, dict):
        iFile = 0
        section = 'sino_file_' + str(iFile)
        while section in dict:
            index = 0
            info = dict[section]
            sino_url = 'sino_url_' + str(index)
            while sino_url in dict[section]:
                fileUrl = info[sino_url]
                if not type(fileUrl) is str:
                    fileUrl = str(fileUrl.tostring().decode())

                _url = url.DataUrl(path=fileUrl)
                dataNameID = 'data_name_' + str(index)
                name = info[dataNameID]
                if hasattr(name, 'astype'):
                    name = str(name.astype('U13'))
                _fileInfo = structs.MatrixFileInfo(file_path=_url.file_path(),
                                                   data_path=_url.data_path())
                self.addSinogram(sinogram=structs.TxSinogram(name=name,
                                                             fileInfo=_fileInfo))
                index += 1
                sino_url = 'sino_url_' + str(index)

            iFile += 1
            section = 'sino_file_' + str(iFile)

    def __eq__(self, other):
        return _ItReconsConfig.__eq__(self, other) and \
               self.getSinogramsFileDict() == other.getSinogramsFileDict()

    def checkSinoType(self, sino):
        assert(isinstance(sino, structs.Sinogram))

    def getSinogramsFileDict(self):
        """
        return the sinograms as a dictionnary with key the file containing the
        sinograms and values the list of sinograms in this file

        :return: dict
        """
        dd = {}
        for sinogram in self.sinograms:
            assert isinstance(sinogram, structs.Sinogram)
            filePath = 'unknow'
            if sinogram.fileInfo and sinogram.fileInfo.file_path():
                filePath = sinogram.fileInfo.file_path()
            if filePath not in dd:
                dd[filePath] = []
            dd[filePath].append(sinogram)
        import collections
        sortedDict = collections.OrderedDict(sorted(dd.items()))
        return sortedDict

    def fillSinogramNames(self):
        # if no name define, define one for each sinogram
        if self.sinograms is not None:
            index = 0
            for sinogram in self.sinograms:
                if sinogram.name is None:
                    sinogram.name = 'sinogram_' + str(index)
                    index += 1


class TxConfig(_SimpleSinoConfig):
    """

    .. note:: Need to get a differentiation between TxConfig and this class
              because tomogui has also to deal with the FBP... and won't have the log
              option and maybe others
    """
    def __init__(self, sinograms=None, computeLog=True, *var, **kw):
        _SimpleSinoConfig.__init__(self, reconsType=self.TX_ID,
                                   sinograms=sinograms, *var, **kw)
        self.computeLog = computeLog

    def _setNormalizationFrmDict(self, dict):
        _SimpleSinoConfig._setNormalizationFrmDict(self, dict)
        assert('normalization' in dict)
        info = dict['normalization']
        assert('computeminuslog' in info)
        self.computeLog = info['computeminuslog']

    def _getNormalization(self):
        dic = _SimpleSinoConfig._getNormalization(self)
        dic['computeminuslog'] = self.computeLog
        return dic

    def __eq__(self, other):
        return _SimpleSinoConfig.__eq__(self, other) and \
               self.computeLog == other.computeLog


class FluoConfig(_SimpleSinoConfig):
    """
    configuration to build fluorescence or compton reconstruction algorithm
    """
    def __init__(self, outBeamCalMethod=0, sinoI0=None, absMat=None,
                 isAbsMatASinogram=False, detector=None, materials=None,
                 sinograms=None, e0=1, *var, **kw):
        _SimpleSinoConfig.__init__(self, reconsType=self.FLUO_ID, *var, **kw)
        assert(absMat is None or isinstance(absMat, structs.AbsMatrix))
        self.absMat = absMat
        if self.absMat is None:
            self.absMat = structs.AbsMatrix(fileInfo=None)

        # TODO: check: not sure I0 is used
        self.sinoI0 = sinoI0
        self.isAbsMatASinogram = isAbsMatASinogram

        assert(materials is None or isinstance(materials, structs.Materials))
        self.materials = materials
        if self.materials is None:
            self.materials = structs.Materials()
        self.detector = detector
        self.outBeamCalMethod = outBeamCalMethod
        self.E0 = e0

    def _defaultSinoType(self):
        return structs.FluoSino

    def _getDataSourceHeader(self):
        return 'data_source_fluo'

    def toDict(self):
        dic = _SimpleSinoConfig.toDict(self)
        dic['detector_setup'] = self._getDetector()
        return dic

    def _sinogramsToDict(self, dic):
        # deal with all fluorescence sinogram
        sinogramsDict = self.getSinogramsFileDict()
        for iSinoFile, sinoFile in enumerate(sinogramsDict):
            section = 'fluo_sino_file_' + str(iSinoFile)
            dic[section] = {}
            indexSino = 0
            for sino in sinogramsDict[sinoFile]:
                fluoSinoUrlID = 'fluo_sino_url_' + str(indexSino)
                dic[section][fluoSinoUrlID] = sino.fileInfo.path()

                dataNameID = 'data_name_' + str(indexSino)
                dic[section][dataNameID] = sino.name

                dataPhysElmtID = 'data_physical_element_' + str(indexSino)
                dic[section][dataPhysElmtID] = sino.physElmt

                EFID = 'ef_' + str(indexSino)
                dic[section][EFID] = sino.EF

                selfAbsMatID = 'self_absorption_file_' + str(indexSino)

                if sino.selfAbsMat is None:
                    dic[section][selfAbsMatID] = ''
                else:
                    assert(isinstance(sino.selfAbsMat, structs.AbsMatrix))
                    assert(sino.selfAbsMat.fileInfo is not None)
                    dic[section][selfAbsMatID] = sino.selfAbsMat.fileInfo.path()

                indexSino += 1

        return dic

    def _fromDict(self, dict):
        _SimpleSinoConfig._fromDict(self, dict)
        assert('detector_setup' in dict)
        info = dict['detector_setup']
        assert('det_width' in info)
        assert('det_pos_x' in info)
        assert('det_pos_y' in info)
        assert('det_pos_z' in info)

        self.detector = structs.Detector(x=info['det_pos_x'],
                                         y=info['det_pos_y'],
                                         z=info['det_pos_z'],
                                         width=info['det_width'])

        # deal with data_source_fluo
        assert('data_source_fluo' in dict)
        info = dict['data_source_fluo']
        self.isAbsMatASinogram = bool(info['absorption_file_is_a_sinogram'])
        absMat = info['absorption_file']
        info_file_abs_mat = _getMatrixFileInfo(absMat)

        if fileInfo.MatrixFileInfo._equalOrNone(info_file_abs_mat, None):
            self.absMat = structs.AbsMatrix()
        else:
            self.absMat = structs.AbsMatrix(fileInfo=info_file_abs_mat)

        materialCompoFile = info['samp_composition_file']
        info_file_comp = _getMatrixFileInfo(materialCompoFile)

        materialsFile = info['materials_file']
        info_file_mat_dic = _getDictFileInfo(materialsFile)

        self.materials = structs.Materials(
            materials=structs.MaterialsDic(fileInfo=info_file_mat_dic),
            matComposition=structs.MatComposition(fileInfo=info_file_comp))

        return self

    def _sinogramsFromDict(self, dict):
        """
        Create list(struct.FluoSino) from dictionnary containing the
        configuration

        :param dict: configuration dictionnary
        """
        iFile = 0
        section = 'fluo_sino_file_' + str(iFile)
        while section in dict:
            info = dict[section]
            indexSino = 0

            sinoUrlID = 'fluo_sino_url_' + str(indexSino)
            while sinoUrlID in info:
                dataNameID = 'data_name_' + str(indexSino)
                dataPhysElmtID = 'data_physical_element_' + str(indexSino)
                EFID = 'ef_' + str(indexSino)
                selfAbsMatID = 'self_absorption_file_' + str(indexSino)
                # deal with sinogram
                sinoFile = info[sinoUrlID]
                file_info = _getMatrixFileInfo(sinoFile)

                # deal with selfAbsMat
                selfAbsMatFile = info[selfAbsMatID] if selfAbsMatID in info else None
                absMatInfo = _getMatrixFileInfo(selfAbsMatFile)

                if fileInfo.MatrixFileInfo._equalOrNone(absMatInfo, None):
                    selfAbsMat = None
                else:
                    selfAbsMat = structs.SelfAbsMatrix(fileInfo=absMatInfo)
                    assert selfAbsMat.fileInfo.is_valid()

                name = info[dataNameID]
                if hasattr(name, 'astype'):
                    name = str(name.astype('U13'))
                physElmt = info[dataPhysElmtID]

                if hasattr(physElmt, 'astype'):
                    physElmt = str(physElmt.astype('U13'))
                # create the fluorescence sinogram
                fluoSino = structs.FluoSino(name=name, fileInfo=file_info,
                                            physElmt=physElmt, ef=info[EFID],
                                            selfAbsMat=selfAbsMat)
                try:
                    fluoSino.loadData()
                except:
                    pass
                self.addSinogram(fluoSino)
                indexSino += 1
                sinoUrlID = 'fluo_sino_url_' + str(indexSino)

            iFile += 1
            section = 'fluo_sino_file_' + str(iFile)

    def _getDataSourceInfo(self):
        assert(type(self.isAbsMatASinogram) is bool)
        compoFile = matFile = None
        if self.materials is not None:
            if(self.materials.matComposition is not None and
                       self.materials.matComposition.data is not None):
                if fileInfo.FreeARTFileInfo._equalOrNone(self.materials.matComposition.fileInfo, None) is False:
                    compoFile = self.materials.matComposition.fileInfo.path()
                    assert (isinstance(self.materials.matComposition, structs.MatComposition))

            if(self.materials.materials is not None and
                       self.materials.materials.data is not None and
                       len(self.materials.materials.data) > 0):
                if fileInfo.FreeARTFileInfo._equalOrNone(self.materials.materials.fileInfo, None) is False:
                    assert(isinstance(self.materials.materials, structs.MaterialsDic))
                    assert(isinstance(self.materials.materials.fileInfo, fileInfo.FreeARTFileInfo))
                    matFile = self.materials.materials.fileInfo.path()

        absMatFile = ''
        if(self.absMat is not None and
               fileInfo.FreeARTFileInfo._equalOrNone(self.absMat.fileInfo, None) is False):
            absMatFile = self.absMat.fileInfo.path()

        return {
            'absorption_file_is_a_sinogram': self.isAbsMatASinogram,
            'absorption_file': absMatFile,
            'samp_composition_file': compoFile or '',
            'materials_file': matFile or '',
        }

    def _getReconsProp(self):
        dic = _ItReconsConfig._getReconsProp(self)
        dic['outgoing_bean_calculation_method'] = self.outBeamCalMethod or 0
        dic['e0'] = self.E0
        return dic

    def _setReconsPropFrmDict(self, dict):
        _ItReconsConfig._setReconsPropFrmDict(self, dict)
        assert('outgoing_bean_calculation_method' in dict['reconstruction_properties'])
        self.outBeamCalMethod = dict['reconstruction_properties']['outgoing_bean_calculation_method']
        assert('e0' in dict['reconstruction_properties'])
        self.E0 = dict['reconstruction_properties']['e0']

    def _getDetector(self):
        if self.detector is None:
            return structs.Detector(x=None,
                                    y=None,
                                    z=None,
                                    width=None).toDict()
        else:
            return self.detector.toDict()

    def setFileInfoToH5File(self, refFile, save=False):
        """
        Change all the fileInfo to point ot the given h5 file

        :param refFile: the h5 file to store information
        :param save: if true once the file info updated, save the dataset
        """
        def dealWithSinograms():
            if self.sinograms:
                sinograms = []
                for sino in self.sinograms:
                    sino.loadData()
                    assert sino.data is not None
                    sino.fileInfo = fileInfo.MatrixFileInfo(file_path=refFile,
                                                            data_path=sino.h5defaultPath)
                    assert sino.data is not None

                    if save is True:
                        sino.save()
                        assert sino.data is not None

                    if(sino.selfAbsMat is not None and
                       fileInfo.FreeARTFileInfo._equalOrNone(sino.selfAbsMat.fileInfo, None) is False):
                        sino.selfAbsMat.loadData()
                        sino.selfAbsMat.fileInfo = fileInfo.MatrixFileInfo(file_path=refFile,
                                                                           data_path=sino.selfAbsMat.h5defaultPath)
                        if save is True:
                            sino.selfAbsMat.save()
                sinograms.append(sino)
                self.sinograms = sinograms

        def dealWithMaterials():
            if self.materials:
                if self.materials.materials:
                    if(fileInfo.FreeARTFileInfo._equalOrNone(self.materials.materials.fileInfo, None) is False):
                        if self.materials.materials.data is None:
                            try:
                                self.materials.materials.loadData()
                            except:
                                pass
                        self.materials.materials.fileInfo = fileInfo.MatrixFileInfo(file_path=refFile,
                                                                                    data_path=structs.MaterialsDic.MATERIALS_DICT)
                    if save is True:
                        self.materials.materials.save()
                if self.materials.matComposition is not None:
                    if(fileInfo.FreeARTFileInfo._equalOrNone(self.materials.matComposition.fileInfo, None) is False):
                        if self.materials.matComposition.data is None:
                            try:
                                self.materials.matComposition.loadData()
                            except:
                                pass
                        self.materials.matComposition.fileInfo = fileInfo.MatrixFileInfo(
                                file_path=refFile,
                                data_path=structs.MatComposition.MAT_COMP_DATASET)
                    if save is True:
                        self.materials.matComposition.save()

        def dealWithAbsMat():
            if self.absMat:
                if(fileInfo.FreeARTFileInfo._equalOrNone(self.absMat.fileInfo, None) is False):
                    if self.absMat.data is None:
                        try:
                            self.absMat.loadData()
                        except:
                            pass
                    self.absMat.fileInfo = fileInfo.MatrixFileInfo(file_path=refFile,
                                                                   data_path=self.absMat.h5defaultPath)
                if save is True:
                    self.absMat.save()


        _ItReconsConfig.setFileInfoToH5File(self, refFile, save)
        dealWithSinograms()
        dealWithMaterials()
        dealWithAbsMat()

    def __eq__(self, other):
        return _SimpleSinoConfig.__eq__(self, other) and \
               self.detector == other.detector and \
               self.absMat == other.absMat and \
               self.isAbsMatASinogram == other.isAbsMatASinogram

    def __str__(self):
        l = _ItReconsConfig.__str__(self)
        l += 'outgoing beam calculation meth = %s\n' % self.outBeamCalMethod
        l += str(self.materials)
        return l

    def checkSinoType(self, sino):
        assert(isinstance(sino, structs.FluoSino))


def retrieveInfoFile(file, index):
    """Simple function to get the adapted infoFile from a file and
    an index

    :param file: the file storing the data
    :param index: int or str location of the data in the file
    """
    assert type(file) is str
    if file == '' or file is None or ord(file[0]) == 0:
        return None
    assert(type(file) is str)
    info_file = None
    if file.lower().endswith('.h5') or file.lower().endswith('.hdf5'):
        fp = str(file) + '::' + str(index)
        info_file = fileInfo.MatrixFileInfo(file_path=fp)
    elif file.lower().endswith('.npy') or file.lower().endswith('.dict') or file.lower().endswith('.ini'):
        info_file = fileInfo.MatrixFileInfo(file_path=file)
    elif file.lower().endswith('.edf'):
        if index == '' or index is None:
            _index = 0
        else:
            _index = int(index)
        info_file = fileInfo.MatrixFileInfo(file_path=str(file), data_slice=int(_index))
    else:
        raise ValueError('extension not managed (%s)' % os.path.basename(file))
    return info_file


def _getMatrixFileInfo(_str):
    assert type(_str) in (type(None), str, numpy.ndarray)
    if hasattr(_str, 'tostring'):
        _str = str(_str.tostring().decode())
    if _str in (None, ''):
        return fileInfo.MatrixFileInfo()
    else:
        _url = url.DataUrl(path=_str)
        return fileInfo.MatrixFileInfo(file_path=_url.file_path(),
                                       data_path=_url.data_path())


def _getDictFileInfo(_str):
    assert type(_str) in (type(None), str, numpy.ndarray)
    if hasattr(_str, 'tostring'):
        _str = str(_str.tostring().decode())
    if _str in (None, ''):
        return fileInfo.DictInfo()
    else:
        _url = url.DataUrl(path=_str)
        return fileInfo.DictInfo(file_path=_url.file_path(),
                                 data_path=_url.data_path())



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Benchmark file."""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/07/2016"

import freeart
from freeart.utils import reconstrutils, exampleutils, outgoingrayalgorithm, raypointsmethod

from silx.gui.plot import Plot2D
from silx.gui import qt

import sys

savingToPdf = False

def benchmark_tx_projection(oversamplings, widthsToTest, nbAngle):
    """
    Create a graph for benchmarking the transmission forward projector

    :param oversamplings: the list of oversamplings to test
    :param widthsToTest: the list of phantom width to test
    :param nbAngle: the number of projection to test
    """
    times = {}
    for oversamp in oversamplings :
        bid = "oversamp=" + str(oversamp)
        times[bid] = {}
        for width in widthsToTest :
            times[bid][width] = exampleutils.produce_tx_SheppLogan_sinogram(oversamp, nbAngle, width, save=False)[2]
            print("tx forward projection - execution time for oversamp = " + str(oversamp) + ", width = " + str(width) + " toke : " + str(times[bid][width]) + " s")

    if savingToPdf :
        title= "tx_benchmatk_for_forward_projection_on_" + str(nbAngle) + "_angles"
        plot = plot2D(title=title )
        for curve in times:
            plot.addCurve(widthsToTest, times[curve].values(), legend=curve )

        plot.legendsDockWidget.show()
        plot.saveGraph(filename=title + '.png', fileFormat='png')
        plot.hide()

def benchmark_tx_reconstruction(oversamplings, widthsToTest, nbAngle, nbIter):
    """
    Create a graph for benchmarking the transmission reconstructor

    :param oversamplings: the list of oversamplings to test
    :param widthsToTest: the list of phantom width to test
    :param nbAngle: the number of projection to test
    :param nbIter: the number of iteration to test (one iteration launch nbAngle back projection)
    """    
    times = {}
    for oversamp in oversamplings :
        bid = "oversamp=" + str(oversamp)
        times[bid] = {}
        for width in widthsToTest :
            sino, angles, time = exampleutils.produce_tx_SheppLogan_sinogram(_oversampling=oversamp, _anglesNb=nbAngle, width=width, save=False )
            times[bid][width] = exampleutils.make_tx_reconstruction(sino, angles, _nbIter=nbIter)[1]
            print("tx reconstruction - execution time for oversamp = " + str(oversamp) + ", width = " + str(width) + " toke : " + str(times[bid][width]) + " s")

    if savingToPdf :
        title = "tx_benchmatk_for_reconstruction_on_" + str(nbAngle) + "_angles_and_" + str(nbIter) + "_iteration"
        plot = plot2D(title=title )
        for curve in times:
            plot.addCurve(widthsToTest, times[curve].values(), legend=curve )

        plot.legendsDockWidget.show()
        plot.saveGraph(filename=(title + ".png"), fileFormat="png")
        plot.hide()


def benchmark_fluo_projection(oversamplings, widthsToTest, nbAngle, outgoingraypointsmethods):
    """
    Benchmark the fluo projection

    :param oversamplings: the list of oversamplings to test
    :param widthsToTest: the list of phantom width to test
    :param nbAngle: the number of projection to test   
    :param outgoingraypointsmethods: the list of outgoingrayalgorithm to test  
    """
    times = {}
    for oversamp in oversamplings :
        for obmc in outgoingraypointsmethods:
            bid = "method="+ str(obmc) + "_oversamp=" + str(oversamp)
            times[bid] = {}
            for width in widthsToTest :
                times[bid][width] = exampleutils.produce_fluo_SheppLogan_sinogram(oversamp, nbAngle, detSetup, width, save=False)[4]
                print("fluo projection - execution time for " + str(oversamp) + " oversampling, with = " + str(width) + " toke : " + str(times[bid][width]))

    if savingToPdf :
        for obmc in outgoingraypointsmethods:
            title = "fluo_benchmark_for_projection_method_" + str(obmc) + "_on_" + str(nbAngle) + "_angles_and_"
            plot = plot2D(title=title )
            for curve in times:
                plot.addCurve(widthsToTest, times[curve].values(), legend=curve )

            plot.legendsDockWidget.show()
            plot.saveGraph(filename=(title + ".png"), fileFormat="png")
            plot.hide()


def benchmark_fluo_reconstruction(oversamplings, widthsToTest, nbAngle, outgoingraypointsmethods, nbIter):
    """
    benchmark the fluo reconstruction

    :param oversamplings: the list of oversamplings to test
    :param widthsToTest: the list of phantom width to test
    :param nbAngle: the number of projection to test    
    :param outgoingraypointsmethods: the list of outgoingrayalgorithm to test  
    :param nbIter: the number of iteration to test (one iteration launch nbAngle back projection)
    """    
    times = {}
    for oversamp in oversamplings :
        for obmc in outgoingraypointsmethods:
            bid = "method="+ str(obmc) + "_oversamp=" + str(oversamp)
            times[bid] = {}
            for width in widthsToTest :
                sinogram, angles, absMat, selfAbsMat, time = exampleutils.produce_fluo_SheppLogan_sinogram(oversamp, nbAngle, detSetup, width, save=False)
                times[bid][width] = exampleutils.make_fluo_reconstruction(sinogram, angles, absMat, selfAbsMat, detSetup, nbIter, oversamp)[1]
                print("fluo reconstruction - execution time for " + str(oversamp) + " oversampling, with = " + str(width) + 
                    ", method = " + str(obmc) + " toke : " + str(times[bid][width]))

    if savingToPdf :
        for obmc in outgoingraypointsmethods:
            title=("fluo_benchmark_for_reconstruction_method_" + str(obmc) + "_on_" + str(nbAngle) + "_angles_and_" + str(nbIter) + "_iteration")
            plot = plot2D(title=title )
            for curve in times:
                plot.addCurve(widthsToTest, times[curve].values(), legend=curve )

            plot.legendsDockWidget.show()
            plot.saveGraph(filename=(title + ".png"), fileFormat="png")
            plot.hide()

if __name__ == "__main__":
    if ('--saving' or '--s') in sys.argv :
        savingToPdf=True

    # benchmark the transmission
    oversamplings = [1, 2, 4, 8]
    widths = [16, 32, 64, 128, 356, 1024]
    methodsBeamsGeneral = [raypointsmethod.withInterpolation ]
    methodsOutgoing = [ outgoingrayalgorithm.rawApproximation,
                outgoingrayalgorithm.createOneRayPerSamplePoint,
                outgoingrayalgorithm.matriceSubdivision]

    if savingToPdf :
        global app  # QApplication must be global to avoid seg fault on quit
        app = qt.QApplication([])

    # TODO : compute mean error
    benchmark_tx_projection(oversamplings, widths, nbAngle = 100)
    benchmark_tx_reconstruction(oversamplings, widths, nbAngle = 100, nbIter=1)

    detPos = (1000, 0, 0)
    detectorRadius = 10 
    detSetup = [(detPos, detectorRadius)]
    
    benchmark_fluo_projection(oversamplings, widths, 100, methodsOutgoing)
    benchmark_fluo_reconstruction(oversamplings, widths, 100, methodsOutgoing, nbIter=1)

    if savingToPdf : 
        app.exec_()  


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "20/10/2016"

import unittest
import numpy

import freeart
from freeart.unitsystem import metricsystem
from freeart.utils import genph

import logging
_logger = logging.getLogger("unitests")

class testTxVsFluo(unittest.TestCase):
    """A simple unit tests to  make sure that if the interaction probability and the self absorption is null 
    then we found back the same stuff for compton and transmission"""

    detectorAngles = [0.0, 60*numpy.pi/180.0, numpy.pi, -numpy.pi/2.0, 1.3*numpy.pi]
    detectorRadius = 1.0
    nbAngles = 20

    def setUp(self):
        phGenerator = genph.PhantomGenerator()
        self.sheppLogan_phantom = phGenerator.get2DPhantomSheppLogan(56)
        self.sheppLogan_phantom.shape = (self.sheppLogan_phantom.shape[0], self.sheppLogan_phantom.shape[1], 1)

        self.probaInteraction = self.sheppLogan_phantom.copy()
        self.probaInteraction[self.probaInteraction!=0.0] = 1.0
        self.selfAbsMat = numpy.zeros(self.sheppLogan_phantom.shape)

    def phantomFluoSinogram(self, withSA, detSetup):
        # define reconstruction parameters
        al = freeart.FluoFwdProjection(
            phMatr           = self.probaInteraction, 
            expSetUp         = detSetup,
            absorpMatr       = self.sheppLogan_phantom,
            selfAbsorpMatrix = self.selfAbsMat,
            angleList        = None, 
            minAngle         = 0, 
            maxAngle         = numpy.pi, 
            anglesNb         = self.nbAngles )

        al.setOverSampling(2)
        al.turnOffSolidAngle(withSA)
        # create the sinogram
        sino, angles = al.makeSinogram()  
        return sino

    def testPhantomReconstruction_withoutSA(self):
        """
        Test that all generated sinogram with a null self absorption is equal no matter the detector position
        """
        sinograms = {}
        for detAngle in self.detectorAngles:
            detPos = (numpy.sin(detAngle)*1000.0, numpy.cos(detAngle)*1000.0, 0.0)
            detSetup = [(detPos, self.detectorRadius)]

            sinograms[detAngle]=self.phantomFluoSinogram(True, detSetup)

        ref = sinograms[0.0]
        for detAngle in self.detectorAngles:
            numpy.testing.assert_allclose(ref, sinograms[detAngle])

    def testPhantomReconstruction_withSA(self):
        """
        Test that all generated sinogram with a null self absorption is equal no matter the detector position
        """
        sinograms = {}
        for detAngle in self.detectorAngles:
            detPos = (numpy.sin(detAngle)*10e5, numpy.cos(detAngle)*10e5, 0.0)
            detSetup = [(detPos, self.detectorRadius)]

            sinograms[detAngle]=self.phantomFluoSinogram(False, detSetup)

        ref = sinograms[0.0]
        for detAngle in self.detectorAngles:
            numpy.testing.assert_allclose(ref, sinograms[detAngle], rtol=5e-4)


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testTxVsFluo))
    return test_suite

if __name__ == '__main__' and __package__ is None :
    unittest.main(defaultTest="suite")


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "29/04/2016"


import logging
import os
import unittest


logger = logging.getLogger(__name__)

from freeart.interpreter.test import suite as test_freeart_suite
from freeart.utils.test import suite as test_utils_suite
from freeart.configuration.test import suite as test_config_suite
from .test_version import suite as test_version_suite
from .test_units import suite as test_units_suite


def suite(loader=None, pattern='test*.py'):
    test_dir = os.path.dirname(__file__)
    if loader is None:
        loader = unittest.TestLoader()

    test_suite = unittest.TestSuite()
    test_suite.addTest(test_freeart_suite())
    test_suite.addTest(test_utils_suite())
    test_suite.addTest(test_version_suite())
    test_suite.addTest(test_units_suite())
    test_suite.addTest(test_config_suite())
    return test_suite


def run_tests():
    """Run test complete test_suite"""
    runner = unittest.TextTestRunner()
    if not runner.run(suite()).wasSuccessful():
        print("Test suite failed")
        return 1
    else:
        print("Test suite succeeded")
        return 0



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Contains the fileInfo classes. They allow the definition of information
to retrieve data from .h5 and .edf files
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "26/10/2017"


from silx.io import url, utils
from silx.io import dictdump
import numpy
import logging

logger = logging.getLogger(__name__)


class FreeARTFileInfo(url.DataUrl):
    """
    class defining potential information for a file sinogram contained in
    a file. And a load function to retrieve the data associated with this
    information

    Now based on the silx.io.url.DataUrl
    """
    def __init__(self, file_path=None, data_path=None, data_slice=None):
        if file_path:
            if len(file_path) == 1 and ord(file_path) == 0:
                file_path = ''
            extension = file_path.split('.')[-1].lower()
        else:
            extension = None

        if extension == 'edf':
            scheme = 'fabio'
        elif extension in ('h5', 'hdf5', 'nx', 'nxs'):
            scheme = 'silx'
        else:
            scheme = None
        _data_slice = data_slice
        if isinstance(_data_slice, (int, float)):
            _data_slice = (_data_slice, )
        url.DataUrl.__init__(self, path=None, file_path=file_path,
                             data_path=data_path, data_slice=_data_slice,
                             scheme=scheme)

    @staticmethod
    def _equalOrNone(mfi1, mfi2):
        """This is not a strict comparison between self and other but this is
        needed, especially for unit test in the case an empty filePath is equal
        to a MatrixFileInfo set to None.
        :warning: Direct inheritance from url.DataUrl.__str__ is not possible
        because enter in conflict with the url.DataUrl source code.
        """
        if mfi1 is not None and mfi2 is None:
            return mfi1.file_path() in (None, '')
        elif mfi2 is not None and mfi1 is None:
            return mfi2.file_path() in (None, '')
        else:
            return mfi1 == mfi2

    def load(self):
        raise NotImplementedError()


class MatrixFileInfo(FreeARTFileInfo):
    def load(self):
        """
        load the data. First from the file set. If None register in infoFile
        then it will use the reffile.

        :return: the data contained
        """
        if self.is_valid() is False:
            err = '%s is not a valid url, cannot load data associated' % self.path
            raise ValueError(err)
        elif self.path() is None or self.file_path() in ("", None):
            return None
        else:
            if self.file_path().lower().endswith('.npy'):
                return numpy.load(self.file_path())
            else:
                try:
                    return utils.get_data(self.path())
                except ValueError as e:
                    # TODO: set it back
                    # logger.error(e)
                    return None


class DictInfo(FreeARTFileInfo):
    def load(self):
        """
        load the data. First from the file set. If None register in infoFile
        then it will use the reffile.

        :return: the data contained
        """
        if self.is_valid() is False:
            err = '%s is not a valid url, cannot load data associated' % self.path
            raise ValueError(err)
        elif self.path() is None or self.file_path() in ("", None):
            return None
        else:
            extension = self.file_path().lower().split('.')[-1]
            if extension in ('h5', 'hdf5', 'hdf', 'nx', 'nxs'):
                return dictdump.h5todict(h5file=self.file_path(),
                                         path=self.data_path())
            elif extension in ('cfg', 'ini'):
                return dictdump.load(self.file_path())
            else:
                logger.error('Extension type %s is not managed by freeart' % extension)
                return None



from freeart.configuration import fileInfo, structs, config, h5ConfigIO
import numpy


def generateTxAllIntegrated():
    outFileTx = 'txReconsIntegrated.h5'

    fp = structs.TxSinogram.TX_SINOGRAM_DATASET
    sinogram = structs.TxSinogram(
        fileInfo=fileInfo.H5MatrixFileInfo(filePath=fp),
        data=numpy.arange(100).reshape(10, 10))
    sinogram.saveTo(outFileTx)

    fp = structs.I0Sinogram.I0_DATASET
    infoFileI0 = fileInfo.H5MatrixFileInfo(filePath=fp)
    I0 = structs.I0Sinogram(data=numpy.arange(100).reshape(10, 10),
                            fileInfo=infoFileI0)
    I0.saveTo(outFileTx)
    txConf = config.TxConfig(sinogram=sinogram,
                             projections='2:6',
                             minAngle=0,
                             maxAngle=1.2,
                             dampingFactor=1.,
                             I0=I0)

    writer = h5ConfigIO.H5ConfigWriter()
    writer.write(outFileTx, txConf)

def generateTxAllOut():
    outFileTx = 'txReconsExternal.h5'

    sinogram = structs.TxSinogram(
        fileInfo=fileInfo.EDFMatrixFileInfo(filePath='sinoTx.edf'),
        data=numpy.arange(100).reshape(10, 10))
    sinogram.save()

    infoFileI0 = fileInfo.EDFMatrixFileInfo(filePath='I0.edf')
    I0 = structs.I0Sinogram(data=numpy.arange(100).reshape(10, 10),
                            fileInfo=infoFileI0)
    I0.save()
    txConf = config.TxConfig(sinogram=sinogram,
                             projections='2:6',
                             minAngle=0,
                             maxAngle=1.2,
                             dampingFactor=1.,
                             I0=I0)

    writer = h5ConfigIO.H5ConfigWriter()
    writer.write(outFileTx, txConf)

if __name__ == '__main__':
    generateTxAllIntegrated()
    generateTxAllOut()


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""test that the TxConfig and FluoConfig class are correct
"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"

import unittest
from .. import config, iniConfigIO, h5ConfigIO, structs, fileInfo
import tempfile
import numpy
import os
import shutil
from freeart.utils import h5utils, reconstrutils
from silx.io import configdict
from silx.io import url
from collections import OrderedDict
import silx
from silx.io import dictdump
from silx.utils.testutils import ParametricTestCase


class TestIniTx(unittest.TestCase):
    """Make sure we can read and write correctly tx configuration to the ini
    format"""
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        txSinoFile = os.path.join(self.tempdir, 'txSino.edf')

        _fileInfo = fileInfo.MatrixFileInfo(file_path=txSinoFile)
        assert _fileInfo.is_valid()
        sino = structs.TxSinogram(data=numpy.arange(100).reshape(10, 10),
                                  fileInfo=_fileInfo)
        sino.save()
        sino.loadData()
        self.txConf = config.TxConfig(
            sinograms=(sino, ),
            projections='2:6',
            minAngle=0,
            maxAngle=1.2,
            dampingFactor=1.)

        self.outFileTx = os.path.join(self.tempdir, 'tx.cfg')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def testReadWrite(self):
        writer = iniConfigIO.IniConfigWriter()
        writer.write(filePath=self.outFileTx,
                     reconsConfiguration=self.txConf)
        reader = iniConfigIO.IniConfigReader()
        confTxLoaded = reader.read(self.outFileTx)
        self.assertTrue(isinstance(confTxLoaded, config.TxConfig))
        self.assertTrue(confTxLoaded == self.txConf)


class TestIniFluo(unittest.TestCase):
    """Make sure we can read and write correctly fluo configuration to the ini
    format"""
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        fluoSinoFile = os.path.join(self.tempdir, 'fluoSino.edf')

        fluoSino = structs.FluoSino(
            fileInfo=fileInfo.MatrixFileInfo(file_path=fluoSinoFile),
            data=numpy.arange(100).reshape(10, 10),
            name='mySinogram',
            physElmt='O',
            ef=12,
            selfAbsMat=None)
        fluoSino.save()
        self.assertTrue(fluoSino.fileInfo.is_valid())

        detector = structs.Detector(x=1, y=2, z=3, width=0.1)
        absMat = structs.AbsMatrix(
            data=numpy.arange(100).reshape(10, 10),
            fileInfo=fileInfo.MatrixFileInfo(
                    file_path=os.path.join(self.tempdir, 'absMat.edf'))
        )
        assert type(absMat.fileInfo) is structs.MatrixFileInfo
        absMat.save()
        absMat.loadData()
        self.assertTrue(absMat.fileInfo.is_valid())

        self.fluoConf = config.FluoConfig(outBeamCalMethod=None,
                                          sinoI0=None,
                                          absMat=absMat,
                                          isAbsMatASinogram=True,
                                          detector=detector,
                                          minAngle=0,
                                          maxAngle=1.2,
                                          dampingFactor=1.,
                                          centerOfRotation=6.,
                                          acquiInverted=True)
        self.fluoConf.addSinogram(fluoSino)
        self.outFileFluo = os.path.join(self.tempdir, 'fluo.cfg')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def testReadWrite(self):
        writer = iniConfigIO.IniConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = iniConfigIO.IniConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)


class TestH5Tx(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.outFileTx = os.path.join(self.tempdir, 'reconsTx.h5')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def testSaveWithFilePath(self):
        """Make sure a simple tx reconstruction is saved correctly"""
        sinoFile = os.path.join(self.tempdir, 'sino.edf')
        sinogram = structs.TxSinogram(
            data=numpy.arange(100).reshape(10, 10),
            fileInfo=fileInfo.MatrixFileInfo(file_path=sinoFile),
            name='sinogramTest')
        sinogram.save()
        self.txConf = config.TxConfig(sinograms=(sinogram,),
                                      projections='2:6',
                                      minAngle=0,
                                      maxAngle=1.2,
                                      dampingFactor=1.,
                                      I0=1.0)

        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(self.outFileTx, self.txConf)

        reader = h5ConfigIO.H5ConfigReader()
        confLoaded = reader.read(self.outFileTx)
        self.assertTrue(self.txConf == confLoaded)

    def testSaveWithI0AsANumpyArray(self):
        """Make sure a tx reconstruction using a sinogram to be normalized
        is saved correctly"""
        sinogram = structs.TxSinogram(
            fileInfo=fileInfo.MatrixFileInfo(
                file_path=self.outFileTx,
                data_path='data2/test/sinogramDiffPlace'
            ),
            data=numpy.arange(100).reshape(10, 10),
            name='sinotest'
        )
        sinogram.save()
        self.assertTrue(sinogram.fileInfo.is_valid())

        infoFileI0 = fileInfo.MatrixFileInfo(
            file_path=os.path.join(self.tempdir, 'I0.edf'),
            data_slice=32)
        I0 = structs.I0Sinogram(data=numpy.arange(100).reshape(10, 10),
                                fileInfo=infoFileI0)
        I0.save()
        self.assertTrue(infoFileI0.is_valid())
        self.txConf = config.TxConfig(sinograms=(sinogram, ),
                                      projections='2:6',
                                      minAngle=0,
                                      maxAngle=1.2,
                                      dampingFactor=1.,
                                      I0=I0)
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(self.outFileTx, self.txConf)
        reader = h5ConfigIO.H5ConfigReader()
        confLoaded = reader.read(self.outFileTx)
        self.assertTrue(self.txConf.I0 == confLoaded.I0)
        self.assertTrue(self.txConf == confLoaded)


class TestH5Fluo(ParametricTestCase):
    """Make sure we can read and write correctly fluo configuration to the ini
    format"""

    NBFluoSino = 5

    def _addFluoSinograms(self, fileExtension, addSelfAbsMat=False,
                          selfAbsMat=None):
        def createFileInfo(fileExtension, name, data):
            if fileExtension is None:
                # TODO : try with official one
                # in this case all data are stored in the same .h5 file
                sourceFile = None
                path = "data/" + name + str(iSino)
                file_info = fileInfo.MatrixFileInfo(file_path=self.outFileFluo,
                                                    data_path=path)
                sourceFile = self.outFileFluo

                h5utils.createH5WithDataSet(url=file_info,
                                            data=dataSino,
                                            mode='a')

            elif fileExtension in ('.h5', '.hdf5'):
                # generate a new h5 file
                fileName = name + 'File' + str(iSino) + '.h5'
                sourceFile = os.path.join(self.tempdir, fileName)

                path = "data/" + name + str(iSino)
                _url = url.DataUrl(file_path=fileName, data_path=path)
                h5utils.createH5WithDataSet(url=_url,
                                            data=dataSino,
                                            mode='a')
                file_info = fileInfo.MatrixFileInfo(file_path=_url.file_path(),
                                                    data_path=_url.data_path())
            elif fileExtension.lower() == '.edf':
                fileName = name + 'File' + str(iSino) + '.edf'
                sourceFile = os.path.join(self.tempdir, fileName)

                file_info = fileInfo.MatrixFileInfo(file_path=sourceFile)
                reconstrutils.saveMatrix(data=dataSino, fileName=sourceFile)
            else:
                raise NotImplementedError('Not managed for other types')

            return file_info, sourceFile

        for iSino in range(self.NBFluoSino):
            dataSino = numpy.arange(100).reshape(10, 10)
            selfAbsMat = None

            fileInfoSino, sourceFileSino = createFileInfo(fileExtension,
                                                          'fluoSino',
                                                          dataSino)
            if addSelfAbsMat is True:
                dataSelfAbsMat = dataSino
                fileInfoSAM, sourceFileSAM = createFileInfo(fileExtension,
                                                            'selfAbsMat',
                                                            dataSelfAbsMat)
                selfAbsMat = structs.SelfAbsMatrix(fileInfo=fileInfoSAM,
                                                   data=dataSelfAbsMat)
            sinogram = structs.FluoSino(fileInfo=fileInfoSino,
                                        data=dataSino,
                                        name='mySinogram',
                                        physElmt='O',
                                        ef=12,
                                        selfAbsMat=selfAbsMat)
            self.fluoConf.addSinogram(sinogram)

    def _addMaterials(self, fileExtension, filePath):
        composition = numpy.arange(100).reshape(10, 10)
        materials = {
            'mat1':
                OrderedDict({
                'CompoundList': ['H', 'O'],
                'CompoundFraction': [0.42, 0.36],
                'Density': 1.0,
                'Thickness': 1.0,
                'Comment': 'wothout'
                }),
            'mat2':
                OrderedDict({
                'CompoundList': ['Y', 'O'],
                'CompoundFraction': [0.42, 0.0],
                'Density': 0.2,
                'Thickness': 0.5,
                'Comment': ''
            }),
        }
        if fileExtension is None:
            # save composition
            _url = url.DataUrl(file_path=filePath,
                               data_path=structs.MatComposition.MAT_COMP_DATASET)
            h5utils.createH5WithDataSet(url=_url,
                                        data=composition,
                                        mode='a')

            dictdump.dicttoh5(h5file=filePath,
                              treedict=materials,
                              h5path=structs.MaterialsDic.MATERIALS_DICT.lstrip("::"),
                              mode='a')
            file_info_dict = fileInfo.MatrixFileInfo(file_path=filePath,
                                                     data_path=structs.MaterialsDic.MATERIALS_DICT)
            file_info_comp = fileInfo.MatrixFileInfo(file_path=filePath,
                                                     data_path=structs.MatComposition.MAT_COMP_DATASET)
        elif fileExtension == '.h5':
            # Then both materials and materials composition are stored in the same .h5
            h5File = os.path.join(self.tempdir, 'matCompo.h5')
            path = "/data/matCompositions"
            _url = url.DataUrl(file_path=filePath, data_path=path)
            h5utils.createH5WithDataSet(url=_url,
                                        data=composition)
            file_info_comp = fileInfo.MatrixFileInfo(file_path=h5File,
                                                     data_path=path)

            h5File = os.path.join(self.tempdir, 'matDict.h5')
            path = '/materials'
            dictdump.dicttoh5(h5file=h5File,
                              treedict=materials,
                              h5path=path)

            file_info_dict = fileInfo.DictInfo(file_path=h5File,
                                               data_path=path)

        elif fileExtension == '.npy':
            npzFile = os.path.join(self.tempdir, 'matCompo.npy')
            file_info_comp = fileInfo.MatrixFileInfo(npzFile)
            numpy.save(npzFile, composition)

            dicFile = os.path.join(self.tempdir, 'matDict.dict.ini')
            tmp = configdict.ConfigDict(materials)
            tmp.write(dicFile)

            assert os.path.isfile(dicFile)
            p = dictdump.load(dicFile)
            assert(p == tmp)
            file_info_dict = fileInfo.MatrixFileInfo(dicFile)
        else:
            raise ValueError('Extansion not managed')

        self.fluoConf.materials = structs.Materials(
            materials=structs.MaterialsDic(fileInfo=file_info_dict,
                                           data=materials),
            matComposition=structs.MatComposition(fileInfo=file_info_comp,
                                                  data=composition)
        )
        self.fluoConf.materials.save()

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.outFileFluo = os.path.join(self.tempdir, 'fluo.h5')

        detector = structs.Detector(x=1, y=2, z=3, width=0.1)

        self.fluoConf = config.FluoConfig(outBeamCalMethod=None,
                                          sinoI0=None,
                                          absMat=None,
                                          isAbsMatASinogram=True,
                                          detector=detector,
                                          minAngle=0,
                                          maxAngle=1.2,
                                          dampingFactor=1.,
                                          centerOfRotation=6.,
                                          acquiInverted=True)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def testI0intSinosIncludedInH5(self):
        """Make sure the data are correctly write and read if all 
        fluorescence sinogram are stored in the same .h5 file"""
        self._addFluoSinograms(fileExtension=None)
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testI0intSinosNotIncludedInH5ButOtherH5(self):
        """Make sure all the data are correctly write and read if all
        fluorescence sinogram are stored in a different .h5 file"""
        self._addFluoSinograms(fileExtension='.h5')
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)
        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testI0intSinosNotIncludedInH5ButOtherEDF(self):
        """Make sure all the data are correctly write and read if all
        fluorescence sinogram are stored in a .edf file"""
        self._addFluoSinograms(fileExtension='.edf')
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testWithMaterialsInH5(self):
        """Test that can load and save if materials are stored in the .h5 file
        where the configuration is also store"""
        self._addMaterials(fileExtension=None, filePath=self.outFileFluo)
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)
        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))

        # ..warning:: As dictdump is not retrieving the correct types we have to update the
        # materials dict but the loaded one will keep the API of the dict
        mat = {}
        for key in confFluoLoaded.materials.materials.data:
            mat[key] = confFluoLoaded.materials.materials.data[key]
            assert("CompoundFraction" in mat[key])
            assert("CompoundList" in mat[key])
            mat[key]["CompoundFraction"] = list(mat[key]["CompoundFraction"])
            mat[key]["CompoundList"] = list(mat[key]["CompoundList"])

        confFluoLoaded.materials.materials.data = mat
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testWithMaterialsExternalH5(self):
        """Test that can load and save if materials are stored in a different
         .h5 file where the configuration is also store"""
        self._addMaterials(fileExtension='.h5', filePath=self.outFileFluo)
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))

        # ..warning:: As dictdump is not retrieving the correct types we have to update the
        # materials dict but the loaded one will keep the API of the dict
        mat = {}
        for key in confFluoLoaded.materials.materials.data:
            mat[key] = confFluoLoaded.materials.materials.data[key]
            assert("CompoundFraction" in mat[key])
            assert("CompoundList" in mat[key])
            mat[key]["CompoundFraction"] = list(mat[key]["CompoundFraction"])
            mat[key]["CompoundList"] = list(mat[key]["CompoundList"])
        confFluoLoaded.materials.materials.data = mat

        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testWithMaterialsExternalNPY(self):
        """Test that can load and save if materials are stored in a different
         .npy file and a different.dict file"""
        self._addMaterials(fileExtension='.npy', filePath=self.outFileFluo)
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        a= self.fluoConf.materials.materials
        b= confFluoLoaded.materials.materials
        confFluoLoaded.materials.loadData()

        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testWithSelfAbsMatricesIncludedInSameH5(self):
        self._addFluoSinograms(fileExtension=None,
                               addSelfAbsMat=True,
                               selfAbsMat=None)
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testWithSelfAbsMatricesIncludedInDifferentH5(self):
        # TODO : merge all selfAbsMat and fluoSinograms with subTest
        self._addFluoSinograms(fileExtension=None,
                               addSelfAbsMat=True,
                               selfAbsMat='.h5')
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)

    def testWithSelfAbsMatricesIncludedInDifferentEDF(self):
        self._addFluoSinograms(fileExtension=None,
                               addSelfAbsMat=True,
                               selfAbsMat='.edf')
        writer = h5ConfigIO.H5ConfigWriter()
        writer.write(filePath=self.outFileFluo,
                     reconsConfiguration=self.fluoConf)

        reader = h5ConfigIO.H5ConfigReader()
        confFluoLoaded = reader.read(self.outFileFluo)
        self.assertTrue(isinstance(confFluoLoaded, config.FluoConfig))
        self.assertTrue(confFluoLoaded == self.fluoConf)


def suite():
    test_suite = unittest.TestSuite()
    for t in (TestIniTx, TestIniFluo, TestH5Tx, TestH5Fluo):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(t))

    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")
